"""
Wrapper for REDUCE C functions
"""
import ctypes
import io
import logging
import os
import sys
import tempfile
from contextlib import contextmanager

import matplotlib.pyplot as plt
import numpy as np

try:
    from .clib._slitfunc_bd import lib as slitfunclib
    from .clib._slitfunc_2d import lib as slitfunc_2dlib

    from .clib._slitfunc_bd import ffi
except ImportError:
    raise ImportError("Use build_extract.py to compile the C libraries")


c_double = np.ctypeslib.ctypes.c_double
c_int = np.ctypeslib.ctypes.c_int

libc = ctypes.CDLL(None)
c_stdout = ctypes.c_void_p.in_dll(libc, 'stdout')

@contextmanager
def stdout_redirector(stream):
    # The original fd stdout points to. Usually 1 on POSIX systems.
    original_stdout_fd = sys.stdout.fileno()

    def _redirect_stdout(to_fd):
        """Redirect stdout to the given file descriptor."""
        # Flush the C-level buffer stdout
        libc.fflush(c_stdout)
        # Flush and close sys.stdout - also closes the file descriptor (fd)
        sys.stdout.close()
        # Make original_stdout_fd point to the same file as to_fd
        os.dup2(to_fd, original_stdout_fd)
        # Create a new sys.stdout that points to the redirected fd
        sys.stdout = io.TextIOWrapper(os.fdopen(original_stdout_fd, 'wb'))

    # Save a copy of the original stdout fd in saved_stdout_fd
    saved_stdout_fd = os.dup(original_stdout_fd)
    try:
        # Create a temporary file and redirect stdout to it
        tfile = tempfile.TemporaryFile(mode='w+b')
        _redirect_stdout(tfile.fileno())
        # Yield to caller, then redirect stdout back to the saved fd
        yield
        _redirect_stdout(saved_stdout_fd)
        # Copy contents of temporary file to the given stream
        tfile.flush()
        tfile.seek(0, io.SEEK_SET)
        stream.write(tfile.read())
    finally:
        tfile.close()
        os.close(saved_stdout_fd)


def slitfunc(img, ycen, lambda_sp=0, lambda_sf=0.1, osample=1):
    """Decompose image into spectrum and slitfunction

    This is for horizontal straight orders only, for curved orders use slitfunc_curved instead

    Parameters
    ----------
    img : array[n, m]
        image to decompose, should just contain a small part of the overall image
    ycen : array[n]
        traces the center of the order along the image, relative to the center of the image?
    lambda_sp : float, optional
        smoothing parameter of the spectrum (the default is 0, which no smoothing)
    lambda_sf : float, optional
        smoothing parameter of the slitfunction (the default is 0.1, which )
    osample : int, optional
        Subpixel ovsersampling factor (the default is 1, which no oversampling)

    Returns
    -------
    sp, sl, model, unc
        spectrum, slitfunction, model, spectrum uncertainties
    """

    # Convert input to expected datatypes
    lambda_sf = float(lambda_sf)
    lambda_sp = float(lambda_sp)
    osample = int(osample)
    img = np.asanyarray(img)
    ycen = np.asanyarray(ycen)

    # Ensure the datatype and shape of all arrays before sending them to C code
    if not np.issubdtype(img.dtype, np.number):
        raise TypeError(
            "Input image must be a numeric type, but got %s" % str(img.dtype)
        )

    if not np.issubdtype(ycen.dtype, np.number):
        raise TypeError("Ycen must be a numeric type, but got %s" % str(ycen.dtype))

    if img.shape[1] != ycen.size:
        raise ValueError(
            "Image and Ycen shapes are incompatible, got %s and %s"
            % (img.shape, ycen.shape)
        )

    if osample <= 0:
        raise ValueError("Oversample rate must be positive, but got %i" % osample)
    if lambda_sf < 0:
        raise ValueError(
            "Slitfunction smoothing must be positive, but got %f" % lambda_sf
        )
    if lambda_sp < 0:
        raise ValueError("Spectrum smoothing must be positive, but got %f" % lambda_sp)

    # Get some derived values
    nrows, ncols = img.shape
    ny = osample * (nrows + 1) + 1
    ycen = ycen - ycen.astype(int)

    # Prepare all arrays
    # Inital guess for slit function and spectrum
    sp = np.ma.sum(img, axis=0)
    sp = np.require(sp, dtype=c_double, requirements=["C", "A", "W", "O"])

    sl = np.zeros(ny, dtype=c_double)

    mask = ~np.ma.getmaskarray(img)
    mask = np.require(mask, dtype=c_int, requirements=["C", "A", "W", "O"])

    img = np.ma.getdata(img)
    img = np.require(img, dtype=c_double, requirements=["C", "A", "W", "O"])

    pix_unc = np.zeros_like(img)
    pix_unc = np.require(pix_unc, dtype=c_double, requirements=["C", "A", "W", "O"])

    ycen = np.require(ycen, dtype=c_double, requirements=["C", "A", "W", "O"])
    model = np.zeros((nrows, ncols), dtype=c_double)
    unc = np.zeros(ncols, dtype=c_double)

    # Call the C function
    slitfunclib.slit_func_vert(
        ffi.cast("int", ncols),
        ffi.cast("int", nrows),
        ffi.cast("double *", img.ctypes.data),
        ffi.cast("double *", pix_unc.ctypes.data),
        ffi.cast("int *", mask.ctypes.data),
        ffi.cast("double *", ycen.ctypes.data),
        ffi.cast("int", osample),
        ffi.cast("double", lambda_sp),
        ffi.cast("double", lambda_sf),
        ffi.cast("double *", sp.ctypes.data),
        ffi.cast("double *", sl.ctypes.data),
        ffi.cast("double *", model.ctypes.data),
        ffi.cast("double *", unc.ctypes.data),
    )
    mask = ~mask.astype(bool)

    return sp, sl, model, unc, mask


def slitfunc_curved(img, ycen, tilt, shear, lambda_sp=0, lambda_sf=0.1, osample=1):
    """Decompose an image into a spectrum and a slitfunction, image may be curved

    Parameters
    ----------
    img : array[n, m]
        input image
    ycen : array[n]
        traces the center of the order
    shear : array[n]
        tilt of the order along the image ???, set to 0 if order straight
    osample : int, optional
        Subpixel ovsersampling factor (the default is 1, which no oversampling)
    lambda_sp : float, optional
        smoothing factor spectrum (the default is 0, which no smoothing)
    lambda_sl : float, optional
        smoothing factor slitfunction (the default is 0.1, which small)

    Returns
    -------
    sp, sl, model, unc
        spectrum, slitfunction, model, spectrum uncertainties
    """

    # Convert datatypes to expected values
    lambda_sf = float(lambda_sf)
    lambda_sp = float(lambda_sp)
    osample = int(osample)
    img = np.asanyarray(img)
    ycen = np.asanyarray(ycen)

    if np.isscalar(tilt) and np.issubdtype(np.asanyarray(tilt).dtype, np.number):
        tilt = np.full(img.shape[1], tilt, dtype=c_double)
    else:
        tilt = np.asanyarray(tilt)
    if np.isscalar(shear) and np.issubdtype(np.asanyarray(shear).dtype, np.number):
        shear = np.full(img.shape[1], shear, dtype=c_double)
    else:
        shear = np.asanyarray(shear)

    # Check type and dimensions of arrays before sending them to the C code
    if not np.issubdtype(img.dtype, np.number):
        raise TypeError(
            "Input image must be a numeric type, but got %s" % str(img.dtype)
        )
    if not np.issubdtype(ycen.dtype, np.number):
        raise TypeError("Ycen must be a numeric type, but got %s" % str(ycen.dtype))
    if not np.issubdtype(tilt.dtype, np.number):
        raise TypeError("Tilt must be a numeric type, but got %s" % str(tilt.dtype))
    if not np.issubdtype(shear.dtype, np.number):
        raise TypeError("Shear must be a numeric type, but got %s" % str(shear.dtype))

    if img.shape[1] != ycen.size:
        raise ValueError(
            "Image and Ycen shapes are incompatible, got %s and %s"
            % (img.shape, ycen.shape)
        )
    if img.shape[1] != tilt.size:
        raise ValueError(
            "Image and Tilt shapes are incompatible, got %s and %s"
            % (img.shape, tilt.shape)
        )
    if img.shape[1] != shear.size:
        raise ValueError(
            "Image and Shear shapes are incompatible, got %s and %s"
            % (img.shape, shear.shape)
        )

    if osample <= 0:
        raise ValueError("Oversample rate must be positive, but got %i" % osample)
    if lambda_sf < 0:
        raise ValueError(
            "Slitfunction smoothing must be positive, but got %f" % lambda_sf
        )
    if lambda_sp < 0:
        raise ValueError("Spectrum smoothing must be positive, but got %f" % lambda_sp)

    # Retrieve some derived values
    nrows, ncols = img.shape
    ny = osample * (nrows + 1) + 1

    ycen_offset = ycen.astype(c_int)
    ycen = ycen - ycen_offset

    y_lower_lim = nrows // 2

    sl = np.zeros(ny, dtype=c_double)

    # Inital guess for spectrum
    sp = np.ma.sum(img, axis=0)
    sp = np.ma.filled(sp, 0)
    sp = np.require(sp, dtype=c_double, requirements=["C", "A", "W", "O"])

    # Initialize arrays and ensure the correct datatype for C
    mask = ~np.ma.getmaskarray(img)
    mask = np.require(mask, dtype=c_int, requirements=["C", "A", "W", "O"])

    img = np.ma.getdata(img)
    img = np.require(img, dtype=c_double, requirements=["C", "A", "W", "O"])

    pix_unc = np.zeros_like(img)
    pix_unc = np.require(pix_unc, dtype=c_double, requirements=["C", "A", "W", "O"])

    ycen = np.require(ycen, dtype=c_double, requirements=["C", "A", "W", "O"])

    ycen_offset = np.require(
        ycen_offset, dtype=c_int, requirements=["C", "A", "W", "O"]
    )

    tilt = np.ma.filled(tilt, 0)
    tilt = np.require(tilt, dtype=c_double, requirements=["C", "A", "W", "O"])
    # TODO DEBUG dont do this
    shear = np.zeros(ncols)
    # shear = np.ma.filled(shear, 0)
    shear = np.require(shear, dtype=c_double, requirements=["C", "A", "W", "O"])

    model = np.zeros((nrows, ncols), dtype=c_double)
    unc = np.zeros(ncols, dtype=c_double)

    # Call the C function
    # f = io.BytesIO()
    # with stdout_redirector(f):
    slitfunc_2dlib.slit_func_curved(
        ffi.cast("int", ncols),
        ffi.cast("int", nrows),
        ffi.cast("double *", img.ctypes.data),
        ffi.cast("double *", pix_unc.ctypes.data),
        ffi.cast("int *", mask.ctypes.data),
        ffi.cast("double *", ycen.ctypes.data),
        ffi.cast("int *", ycen_offset.ctypes.data),
        ffi.cast("double *", tilt.ctypes.data),
        ffi.cast("double *", shear.ctypes.data),
        ffi.cast("int", y_lower_lim),
        ffi.cast("int", osample),
        ffi.cast("double", lambda_sp),
        ffi.cast("double", lambda_sf),
        ffi.cast("double *", sp.ctypes.data),
        ffi.cast("double *", sl.ctypes.data),
        ffi.cast("double *", model.ctypes.data),
        ffi.cast("double *", unc.ctypes.data),
    )
    # output = f.getvalue().decode("utf-8")
    # if output != "":
    #     img.tofile("debug_img.dat")
    #     pix_unc.tofile("debug_pix_unc.dat")
    #     mask.tofile("debug_mask.dat")
    #     ycen.tofile("debug_ycen.dat")
    #     ycen_offset.tofile("debug_ycen_offset.dat")
    #     tilt.tofile("debug_tilt.dat")
    #     shear.tofile("debug_shear.dat")
    #     print(output)

    mask = ~mask.astype(bool)

    return sp, sl, model, unc, mask



import os.path
import logging
import json
import jsonschema

if int(jsonschema.__version__[0]) < 3:
    logging.warning(f"Jsonschema {jsonschema.__version__} found, but at least 3.0.0 is required to check configuration. Skipping the check.")
    hasJsonSchema = False
else:
    hasJsonSchema = True

def get_configuration_for_instrument(instrument, plot=None):
    local = os.path.dirname(__file__)
    fname = os.path.join(local, "settings", f"settings_{instrument.upper()}.json")

    config = load_config(fname, instrument)

    if plot is not None:
        for key, value in config.items():
            if isinstance(config[key], dict) and "plot" in config[key].keys():
                config[key]["plot"] = plot

    return config

def load_config(configuration, instrument, j=0):
    if configuration is None:
        config = "settings_%s.json" % instrument.upper()
    elif isinstance(configuration, dict):
        if instrument in configuration.keys():
            config = configuration[instrument]
        elif "__instrument__" in configuration.keys() and configuration["__instrument__"] == instrument.upper():
            config = configuration
        else:
            raise KeyError("This configuration is for a different instrument")
    elif isinstance(configuration, list):
        config = configuration[j]
    elif isinstance(configuration, str):
        config = configuration

    if isinstance(config, str):
        if os.path.isfile(config):
            logging.info("Loading configuration from %s", config)
            with open(config) as f:
                config = json.load(f)
        else:
            logging.warning(
                "No configuration found at %s, using default values", config
            )
            config = {}

    # Load general settings
    settings = read_config()
    nparam1 = count_parameters(settings)
    # Update values with given settings
    for key in settings.keys():
        if key in config.keys():
            if isinstance(settings[key], dict):
                settings[key].update(config[key])
            else:
                settings[key] = config[key]
    nparam2 = count_parameters(settings, exclude="instrument")
    if nparam2 > nparam1:
        logging.warning("New parameter(s) in instrument config, Check spelling!")

    # If it doesn't raise an Exception everything is as expected
    validate_config(settings)
    logging.debug("Configuration succesfully validated")

    return settings


def read_config(fname="settings_pyreduce.json"):
    this_dir = os.path.dirname(__file__)
    fname = os.path.join(this_dir, "settings", fname)

    if os.path.exists(fname):
        with open(fname) as file:
            settings = json.load(file)
            return settings
    else:
        raise FileNotFoundError(f"Settings file {fname} not found")


def validate_config(config):
    if not hasJsonSchema:
        # Can't check with old version
        return
    fname = "settings_schema.json"
    this_dir = os.path.dirname(__file__)
    fname = os.path.join(this_dir, "settings", fname)

    with open(fname) as f:
        schema = json.load(f)
    try:
        jsonschema.validate(schema=schema, instance=config)
    except jsonschema.ValidationError as ve:
        logging.error("Configuration failed validation check.\n%s", ve.message)
        raise ve


def count_parameters(config, exclude=()):
    nparam = 0

    for key, value in config.items():
        if isinstance(value, dict) and key not in exclude:
            nparam += count_parameters(value)
        elif key not in exclude:
            nparam += 1

    return nparam



import pytest
import numpy as np

from pyreduce.clipnflip import clipnflip

# clipnflip(image, header, xrange=None, yrange=None, orientation=None)


@pytest.fixture
def image():
    image = np.arange(0, 200).reshape((20, 10))
    return image


@pytest.fixture
def header(image):
    # Default header is set up to change nothing
    # Note that image indices are y and x (and not x and y)
    header = {}
    header["e_amp"] = 1
    header["e_xlo"], header["e_xhi"] = 0, image.shape[1]
    header["e_ylo"], header["e_yhi"] = 0, image.shape[0]
    header["e_orient"] = 0
    return header


def test_nochange(image, header):
    # This should change nothing
    flipped = clipnflip(image, header)

    assert isinstance(flipped, np.ndarray)
    assert flipped.shape[0] == image.shape[0]
    assert flipped.shape[1] == image.shape[1]
    assert np.all(flipped == image)


def test_only_rotate(image, header):
    for orient in [0, 1, 2, 3]:
        flipped = clipnflip(image, header, orientation=orient)

        compare = np.rot90(image, -1 * orient)

        assert isinstance(flipped, np.ndarray)
        assert flipped.shape[0] == compare.shape[0]
        assert flipped.shape[1] == compare.shape[1]
        assert np.all(flipped == compare)


def test_only_clip(image, header):
    # clip x direction
    flipped = clipnflip(image, header, xrange=(5, 8))

    assert isinstance(flipped, np.ndarray)
    assert flipped.shape[0] == 20
    assert flipped.shape[1] == 3
    assert np.all(flipped == image[:, 5:8])

    # clip y direction
    flipped = clipnflip(image, header, yrange=(10, 15))

    assert isinstance(flipped, np.ndarray)
    assert flipped.shape[0] == 5
    assert flipped.shape[1] == 10
    assert np.all(flipped == image[10:15, :])


def test_bad_clipping_range(image, header):
    # Case 1: x completely out of range
    with pytest.raises(IndexError):
        flipped = clipnflip(image, header, xrange=(100, 150))

    # Case 2: x upper limit out of range
    with pytest.raises(IndexError):
        flipped = clipnflip(image, header, xrange=(5, 150))

    # Case 3: x limits inverse
    with pytest.raises(IndexError):
        flipped = clipnflip(image, header, xrange=(10, 0))

    # Case 4: x limits only one column
    with pytest.raises(IndexError):
        flipped = clipnflip(image, header, xrange=(1, 1))

    # Case 1: y completely out of range
    with pytest.raises(IndexError):
        flipped = clipnflip(image, header, yrange=(100, 150))

    # Case 2: y upper limit out of range
    with pytest.raises(IndexError):
        flipped = clipnflip(image, header, yrange=(5, 150))

    # Case 3: y limits inverse
    with pytest.raises(IndexError):
        flipped = clipnflip(image, header, yrange=(10, 0))

    # Case 4: y limits only one column
    with pytest.raises(IndexError):
        flipped = clipnflip(image, header, yrange=(1, 1))



import pytest
import numpy as np

from pyreduce.continuum_normalization import splice_orders


def test_continuum(spec, wave, normflat, order_range):
    spec, sigma = spec
    norm, blaze = normflat
    wave, thar = wave

    spec, wave, blaze, sigma = splice_orders(
        spec, wave, blaze, sigma, scaling=True, plot=False
    )

    assert isinstance(spec, np.ma.masked_array)
    assert isinstance(wave, np.ma.masked_array)
    assert isinstance(blaze, np.ma.masked_array)
    assert isinstance(sigma, np.ma.masked_array)

    assert spec.ndim == wave.ndim == blaze.ndim == sigma.ndim == 2
    assert (
        spec.shape[0]
        == wave.shape[0]
        == blaze.shape[0]
        == sigma.shape[0]
        == order_range[1] - order_range[0]
    )
    assert (
        spec.shape[1]
        == wave.shape[1]
        == blaze.shape[1]
        == sigma.shape[1]
        == norm.shape[1]
    )




import os
import tarfile
import wget


def UVES_HD132205(local_dir=None):
    """Load an example dataset for the UVES instrument

    Parameters
    ----------
    local_dir : str, optional
        directory to save data at (default: "./")

    Returns
    -------
    dataset_dir : str
        directory where the data was saved
    """

    if local_dir is None:
        local_dir = os.path.dirname(__file__)
        local_dir = os.path.join(local_dir, "../")

    # load data if necessary
    target_dir = os.path.join(local_dir, "datasets")
    filename = os.path.join(local_dir, "uves_data.tar.gz")

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    if not os.path.isfile(filename):
        url = "http://www.astro.uu.se/~piskunov/RESEARCH/REDUCE/FTP/reduce_demo.tar.gz"
        wget.download(url, out=filename)

    with tarfile.open(filename) as file:
        file.extractall(path=target_dir)

    return target_dir


if __name__ == "__main__":
    UVES_HD132205()



import pytest
import numpy as np

from pyreduce.estimate_background_scatter import estimate_background_scatter


def test_scatter(flat, orders):
    img, _ = flat
    orders, column_range = orders

    scatter = estimate_background_scatter(
        img, orders, scatter_degree=4, column_range=column_range
    )

    assert isinstance(scatter, np.ndarray)
    assert scatter.ndim == 2
    assert scatter.shape[0] == 5
    assert scatter.shape[1] == 5


def test_simple():
    img = np.full((100, 100), 10.)
    orders = np.array([[0, 25], [0, 50], [0, 75]])

    scatter = estimate_background_scatter(img, orders, scatter_degree=0, plot=False)

    assert isinstance(scatter, np.ndarray)
    assert scatter.ndim == 2
    assert scatter.shape[0] == 1
    assert scatter.shape[1] == 1

    assert np.allclose(scatter[0, 0], 10.)


def test_scatter_degree():
    img = np.full((100, 100), 10.)
    orders = np.full((2, 2), 1.)

    estimate_background_scatter(img, orders, scatter_degree=0)

    with pytest.raises(ValueError):
        estimate_background_scatter(img, orders, scatter_degree=-1)

    with pytest.raises(TypeError):
        estimate_background_scatter(img, orders, scatter_degree=2.5)

    estimate_background_scatter(img, orders, scatter_degree=(2, 2))

    with pytest.raises(ValueError):
        estimate_background_scatter(img, orders, scatter_degree=(1,))

    with pytest.raises(ValueError):
        estimate_background_scatter(img, orders, scatter_degree=(3, 2, 1))

    with pytest.raises(ValueError):
        estimate_background_scatter(img, orders, scatter_degree=(2, -1))

    with pytest.raises(TypeError):
        estimate_background_scatter(img, orders, scatter_degree=(2, 2.5))



"""
Find clusters of pixels with signal
And combine them into continous orders
"""

import logging
from itertools import combinations
from functools import cmp_to_key

import matplotlib.pyplot as plt
import numpy as np
from numpy.polynomial.polynomial import Polynomial

from scipy.ndimage import morphology, label
from scipy.ndimage.filters import gaussian_filter1d, median_filter
from scipy.signal import peak_widths, find_peaks

from .util import polyfit1d


def fit(x, y, deg, regularization=0):
    # order = polyfit1d(y, x, deg, regularization)
    order = Polynomial.fit(y, x, deg=deg, domain=[]).coef[::-1]
    return order


def determine_overlap_rating(xi, yi, xj, yj, mean_cluster_thickness, nrow, ncol, deg=2):
    i_left, i_right = yi.min(), yi.max()
    j_left, j_right = yj.min(), yj.max()

    order_i = fit(xi, yi, deg)
    order_j = fit(xj, yj, deg)

    # Get polynomial points inside cluster limits for each
    y_ii = np.polyval(order_i, np.arange(i_left, i_right))
    y_ij = np.polyval(order_i, np.arange(j_left, j_right))
    y_jj = np.polyval(order_j, np.arange(j_left, j_right))
    y_ji = np.polyval(order_j, np.arange(i_left, i_right))

    # difference of polynomials within each cluster limit
    diff_i = np.abs(y_ii - y_ji)
    diff_j = np.abs(y_ij - y_jj)

    ind_i = np.where((diff_i < mean_cluster_thickness) & (y_ji >= 0) & (y_ji < nrow))
    ind_j = np.where((diff_j < mean_cluster_thickness) & (y_ij >= 0) & (y_ij < nrow))

    overlap = len(ind_i[0]) + len(ind_j[0])
    overlap = overlap / ((i_right - i_left) + (j_right - j_left))
    if i_right < j_left:
        overlap *= 1 - (i_right - j_left) / ncol
    elif j_right < i_left:
        overlap *= 1 - (j_right - i_left) / ncol

    overlap_region = [-1, -1]
    if len(ind_i[0]) > 0:
        overlap_region[0] = np.min(ind_i[0]) + i_left
    if len(ind_j[0]) > 0:
        overlap_region[1] = np.max(ind_j[0]) + j_left

    return overlap, overlap_region


def create_merge_array(x, y, mean_cluster_thickness, nrow, ncol, deg, threshold):
    n_clusters = list(x.keys())
    nmax = len(n_clusters) ** 2
    merge = np.zeros((nmax, 5))
    for k, (i, j) in enumerate(combinations(n_clusters, 2)):
        overlap, region = determine_overlap_rating(
            x[i], y[i], x[j], y[j], mean_cluster_thickness, nrow, ncol, deg=deg
        )
        merge[k] = [i, j, overlap, *region]
    merge = merge[merge[:, 2] > threshold]
    merge = merge[np.argsort(merge[:, 2])[::-1]]
    return merge


def update_merge_array(
    merge, x, y, j, mean_cluster_thickness, nrow, ncol, deg, threshold
):
    j = int(j)
    n_clusters = np.array(list(x.keys()))
    update = []
    for i in n_clusters[n_clusters != j]:
        overlap, region = determine_overlap_rating(
            x[i], y[i], x[j], y[j], mean_cluster_thickness, nrow, ncol, deg=deg
        )
        if overlap <= threshold:
            # no , or little overlap
            continue
        update += [[i, j, overlap, *region]]
    if len(update) == 0:
        return merge
    update = np.array(update)
    merge = np.concatenate((merge, update))
    merge = merge[np.argsort(merge[:, 2])[::-1]]
    return merge


def calculate_mean_cluster_thickness(x, y):
    # Calculate mean cluster thickness
    # TODO optimize
    n_clusters = list(x.keys())
    mean_cluster_thickness = 10
    for cluster in n_clusters:
        # individual columns of this cluster
        columns = np.unique(y[cluster])
        delta = 0
        for col in columns:
            # thickness of the cluster in each column
            tmp = x[cluster][y[cluster] == col]
            delta += np.max(tmp) - np.min(tmp)
        mean_cluster_thickness += delta / len(columns)

    mean_cluster_thickness *= 1.5 / len(n_clusters)
    return mean_cluster_thickness


def delete(i, x, y, merge):
    del x[i], y[i]
    merge = merge[(merge[:, 0] != i) & (merge[:, 1] != i)]
    return x, y, merge


def combine(i, j, x, y, merge, mct, nrow, ncol, deg, threshold):
    # Merge pixels
    y[j] = np.concatenate((y[j], y[i]))
    x[j] = np.concatenate((x[j], x[i]))
    # Delete obsolete data
    x, y, merge = delete(i, x, y, merge)
    merge = merge[(merge[:, 0] != j) & (merge[:, 1] != j)]
    # Update merge array
    merge = update_merge_array(merge, x, y, j, mct, nrow, ncol, deg, threshold)
    return x, y, merge


def merge_clusters(
    img,
    x,
    y,
    n_clusters,
    manual=True,
    deg=2,
    auto_merge_threshold=0.9,
    merge_min_threshold=0.1,
):
    """Merge clusters that belong together

    Parameters
    ----------
    img : array[nrow, ncol]
        the image the order trace is based on
    orders : dict(int, array(float))
        coefficients of polynomial fits to clusters
    x : dict(int, array(int))
        x coordinates of cluster points
    y : dict(int, array(int))
        y coordinates of cluster points
    n_clusters : array(int)
        cluster numbers
    threshold : int, optional
        overlap threshold for merging clusters (the default is 100)
    manual : bool, optional
        if True ask before merging orders

    Returns
    -------
    x : dict(int: array)
        x coordinates of clusters, key=cluster id
    y : dict(int: array)
        y coordinates of clusters, key=cluster id
    n_clusters : int
        number of identified clusters
    """

    nrow, ncol = img.shape
    mct = calculate_mean_cluster_thickness(x, y)

    merge = create_merge_array(x, y, mct, nrow, ncol, deg, merge_min_threshold)

    if manual:
        plt.ion()

    k = 0
    while k < len(merge):
        i, j, overlap, _, _ = merge[k]
        i, j = int(i), int(j)

        if overlap >= auto_merge_threshold and auto_merge_threshold != 1:
            answer = "y"
        elif manual:
            plot_order(i, j, x, y, img, deg, title=f"Probability: {overlap}")
            while True:
                if manual:
                    answer = input("Merge? [y/n]")
                if answer in "ynrg":
                    break
        else:
            answer = "y"

        if answer == "y":
            # just merge automatically
            logging.info("Merging orders %i and %i", i, j)
            x, y, merge = combine(
                i, j, x, y, merge, mct, nrow, ncol, deg, merge_min_threshold
            )
        elif answer == "n":
            k += 1
        elif answer == "r":
            x, y, merge = delete(i, x, y, merge)
        elif answer == "g":
            x, y, merge = delete(j, x, y, merge)

    if manual:
        plt.close()
        plt.ioff()

    n_clusters = list(x.keys())
    return x, y, n_clusters


def fit_polynomials_to_clusters(x, y, clusters, degree, regularization=0):
    """Fits a polynomial of degree opower to points x, y in cluster clusters

    Parameters
    ----------
    x : dict(int: array)
        x coordinates seperated by cluster
    y : dict(int: array)
        y coordinates seperated by cluster
    clusters : list(int)
        cluster labels, equivalent to x.keys() or y.keys()
    degree : int
        degree of polynomial fit
    Returns
    -------
    orders : dict(int, array[degree+1])
        coefficients of polynomial fit for each cluster
    """

    orders = {c: fit(x[c], y[c], degree, regularization) for c in clusters}
    return orders


def plot_orders(im, x, y, clusters, orders, order_range):
    """ Plot orders and image """

    cluster_img = np.zeros(im.shape, dtype=im.dtype)
    for c in clusters:
        cluster_img[x[c], y[c]] = c + 1
    cluster_img = np.ma.masked_array(cluster_img, mask=cluster_img == 0)

    plt.subplot(121)
    bot, top = np.percentile(im, (1, 99))
    plt.imshow(im, origin="lower", vmin=bot, vmax=top)
    plt.title("Input Image + Order polynomials")
    plt.xlabel("x [pixel]")
    plt.ylabel("y [pixel]")
    plt.ylim([0, im.shape[0]])

    if orders is not None:
        for i, order in enumerate(orders):
            x = np.arange(*order_range[i], 1)
            y = np.polyval(order, x)
            plt.plot(x, y)

    plt.subplot(122)
    plt.imshow(cluster_img, cmap=plt.get_cmap("tab20"), origin="upper")
    plt.title("Detected Clusters + Order Polynomials")
    plt.xlabel("x [pixel]")
    plt.ylabel("y [pixel]")

    if orders is not None:
        for i, order in enumerate(orders):
            x = np.arange(*order_range[i], 1)
            y = np.polyval(order, x)
            plt.plot(x, y)

    plt.ylim([0, im.shape[0]])
    plt.show()


def plot_order(i, j, x, y, img, deg, title=""):
    """ Plot a single order """
    _, ncol = img.shape

    order_i = fit(x[i], y[i], deg)
    order_j = fit(x[j], y[j], deg)

    xp = np.arange(ncol)
    yi = np.polyval(order_i, xp)
    yj = np.polyval(order_j, xp)

    xmin = min(np.min(x[i]), np.min(x[j])) - 50
    xmax = max(np.max(x[i]), np.max(x[j])) + 50
    ymin = min(np.min(y[i]), np.min(y[j])) - 50
    ymax = max(np.max(y[i]), np.max(y[j])) + 50

    yymin = min(max(0, ymin), img.shape[0]-2)
    yymax = min(ymax, img.shape[0]-1)
    xxmin = min(max(0, xmin), img.shape[1]-2)
    xxmax = min(xmax, img.shape[1]-1)

    vmin, vmax = np.percentile(img[yymin:yymax, xxmin:xxmax], (5, 95))

    plt.clf()
    plt.title(title)
    plt.imshow(img, vmin=vmin, vmax=vmax)
    plt.plot(xp, yi, "r")
    plt.plot(xp, yj, "g")
    plt.plot(y[i], x[i], "r.")
    plt.plot(y[j], x[j], "g.")
    plt.xlim([ymin, ymax])
    plt.ylim([xmin, xmax])
    plt.show()


def mark_orders(
    im,
    min_cluster=None,
    min_width=None,
    filter_size=None,
    noise=None,
    opower=4,
    border_width=None,
    degree_before_merge=2,
    regularization=0,
    closing_shape=(5, 5),
    opening_shape=(2, 2),
    plot=False,
    manual=True,
    auto_merge_threshold=0.9,
    merge_min_threshold=0.1,
    sigma=2,
):
    """ Identify and trace orders

    Parameters
    ----------
    im : array[nrow, ncol]
        order definition image
    min_cluster : int, optional
        minimum cluster size in pixels (default: 500)
    filter_size : int, optional
        size of the running filter (default: 120)
    noise : float, optional
        noise to filter out (default: 8)
    opower : int, optional
        polynomial degree of the order fit (default: 4)
    border_width : int, optional
        number of pixels at the bottom and top borders of the image to ignore for order tracing (default: 5)
    plot : bool, optional
        wether to plot the final order fits (default: False)
    manual : bool, optional
        wether to manually select clusters to merge (strongly recommended) (default: True)

    Returns
    -------
    orders : array[nord, opower+1]
        order tracing coefficients (in numpy order, i.e. largest exponent first)
    """

    # Convert to signed integer, to avoid underflow problems
    im = np.asanyarray(im)
    im = im.astype(int)

    if filter_size is None:
        col = im[:, im.shape[0] // 2]
        col = median_filter(col, 5)
        threshold = np.percentile(col, 90)
        npeaks = find_peaks(col, height=threshold)[0].size
        filter_size = im.shape[0] // npeaks
        logging.info("Median filter size, estimated: %i", filter_size)
    elif filter_size <= 0:
        raise ValueError(f"Expected filter size > 0, but got {filter_size}")

    if border_width is None:
        # find width of orders, based on central column
        col = im[:, im.shape[0] // 2]
        col = median_filter(col, 5)
        idx = np.argmax(col)
        width = peak_widths(col, [idx])[0][0]
        border_width = int(np.ceil(width))
        logging.info("Image border width, estimated: %i", border_width)
    elif border_width < 0:
        raise ValueError(f"Expected border width > 0, but got {border_width}")

    if min_cluster is None:
        min_cluster = im.shape[1] // 4
        logging.info("Minimum cluster size, estimated: %i", min_cluster)
    elif not np.isscalar(min_cluster):
        raise TypeError(f"Expected scalar minimum cluster size, but got {min_cluster}")

    if min_width is None:
        min_width = 0.25
    if min_width == 0:
        pass
    elif isinstance(min_width, (float, np.floating)):
        min_width = int(min_width * im.shape[0])
        logging.info("Minimum order width, estimated: %i", min_width)

    # blur image along columns, and use the median + blurred + noise as threshold
    blurred = gaussian_filter1d(im, filter_size, axis=0)

    if noise is None:
        tmp = np.abs(blurred.flatten())
        noise = np.percentile(tmp, 5)
        logging.info("Background noise, estimated: %f", noise)
    elif not np.isscalar(noise):
        raise TypeError(f"Expected scalar noise level, but got {noise}")

    threshold = np.ma.median(blurred - im, axis=0)
    mask = im > blurred + noise + np.abs(threshold)
    # remove borders
    if border_width != 0:
        mask[:border_width, :] = mask[-border_width:, :] = False
        mask[:, :border_width] = mask[:, -border_width:] = False
    # remove masked areas with no clusters
    mask = np.ma.filled(mask, fill_value=False)
    # close gaps inbetween clusters
    struct = np.full(closing_shape, 1)
    mask = morphology.binary_closing(mask, struct, border_value=1)
    # remove small lonely clusters
    struct = np.full(opening_shape, 1)
    # struct = morphology.generate_binary_structure(2, 1)
    mask = morphology.binary_opening(mask, struct)

    # label clusters
    clusters, _ = label(mask)

    # remove small clusters
    sizes = np.bincount(clusters.ravel())
    mask_sizes = sizes > min_cluster
    mask_sizes[0] = True  # This is the background, which we don't need to remove
    for i in np.arange(len(sizes))[~mask_sizes]:
        clusters[clusters == i] = 0

    # # Reorganize x, y, clusters into a more convenient "pythonic" format
    # # x, y become dictionaries, with an entry for each order
    # # n is just a list of all orders (ignore cluster == 0)
    n = np.unique(clusters)
    n = n[n != 0]
    x = {i: np.where(clusters == c)[0] for i, c in enumerate(n)}
    y = {i: np.where(clusters == c)[1] for i, c in enumerate(n)}

    def best_fit_degree(x, y):
        L1 = np.sum((np.polyval(np.polyfit(y, x, 1), y) - x) ** 2)
        L2 = np.sum((np.polyval(np.polyfit(y, x, 2), y) - x) ** 2)

        # aic1 = 2 + 2 * np.log(L1) + 4 / (x.size - 2)
        # aic2 = 4 + 2 * np.log(L2) + 12 / (x.size - 3)

        if L1 < L2:
            return 1
        else:
            return 2

    if sigma > 0:
        degree = {i: best_fit_degree(x[i], y[i]) for i in x.keys()}
        bias = {i: np.polyfit(y[i], x[i], deg=degree[i])[-1] for i in x.keys()}
        n = list(x.keys())
        yt = np.concatenate([y[i] for i in n])
        xt = np.concatenate([x[i] - bias[i] for i in n])
        coef = np.polyfit(yt, xt, deg=degree_before_merge)

        res = np.polyval(coef, yt)
        cutoff = sigma * (res - xt).std()

        # DEBUG plot
        # uy = np.unique(yt)
        # mask = np.abs(res - xt) > cutoff
        # plt.plot(yt, xt, ".")
        # plt.plot(yt[mask], xt[mask], "r.")
        # plt.plot(uy, np.polyval(coef, uy))
        # plt.show()
        #

        m = {
            i: np.abs(np.polyval(coef, y[i]) - (x[i] - bias[i])) < cutoff for i in x.keys()
        }

        k = max(x.keys()) + 1
        for i in range(1, k):
            new_img = np.zeros(im.shape, dtype=int)
            new_img[x[i][~m[i]], y[i][~m[i]]] = 1
            clusters, _ = label(new_img)

            x[i] = x[i][m[i]]
            y[i] = y[i][m[i]]
            if len(x[i]) == 0:
                del x[i], y[i]

            nnew = np.max(clusters)
            if nnew != 0:
                xidx, yidx = np.indices(im.shape)
                for j in range(1, nnew + 1):
                    xn = xidx[clusters == j]
                    yn = yidx[clusters == j]
                    if xn.size >= min_cluster:
                        x[k] = xn
                        y[k] = yn
                        k += 1
                # plt.imshow(clusters, origin="lower")
                # plt.show()

    if plot:
        plt.title("Identified clusters")
        plt.xlabel("x [pixel]")
        plt.ylabel("y [pixel]")
        clusters = np.ma.zeros(im.shape, dtype=int)
        for i in x.keys():
            clusters[x[i], y[i]] = i + 1
        clusters[clusters == 0] = np.ma.masked

        plt.imshow(clusters, origin="lower", cmap="prism")
        plt.show()

    # Merge clusters, if there are even any possible mergers left
    x, y, n = merge_clusters(
        im,
        x,
        y,
        n,
        manual=manual,
        deg=degree_before_merge,
        auto_merge_threshold=auto_merge_threshold,
        merge_min_threshold=merge_min_threshold,
    )

    if min_width > 0:
        sizes = {k: v.max() - v.min() for k, v in y.items()}
        mask_sizes = {k: v > min_width for k, v in sizes.items()}
        for k, v in mask_sizes.items():
            if not v:
                del x[k]
                del y[k]
        n = x.keys()

    orders = fit_polynomials_to_clusters(x, y, n, opower)

    # sort orders from bottom to top, using relative position

    def compare(i, j):
        _, xi, i_left, i_right = i
        _, xj, j_left, j_right = j

        if i_right < j_left or j_right < i_left:
            return xi.mean() - xj.mean()

        left = max(i_left, j_left)
        right = min(i_right, j_right)

        return xi[left:right].mean() - xj[left:right].mean()

    xp = np.arange(im.shape[1])
    keys = [(c, np.polyval(orders[c], xp), y[c].min(), y[c].max()) for c in x.keys()]
    keys = sorted(keys, key=cmp_to_key(compare))
    key = [k[0] for k in keys]

    n = np.arange(len(n), dtype=int)
    x = {c: x[key[c]] for c in n}
    y = {c: y[key[c]] for c in n}
    orders = np.array([orders[key[c]] for c in n])

    column_range = np.array([[np.min(y[i]), np.max(y[i]) + 1] for i in n])

    if plot:
        plot_orders(im, x, y, n, orders, column_range)

    return orders, column_range



import pytest
import numpy as np

from pyreduce.util import load_fits
from pyreduce.make_shear import Curvature as CurvatureModule


def test_shear(files, wave, orders, instrument, mode, extension, mask, order_range):
    _, extracted = wave
    orders, column_range = orders

    files = files["curvature"][0]
    original, thead = load_fits(files, instrument, mode, extension, mask=mask)

    module = CurvatureModule(
        orders, column_range=column_range, order_range=order_range, plot=False
    )
    tilt, shear = module.execute(extracted, original)

    assert isinstance(tilt, np.ndarray)
    assert tilt.ndim == 2
    assert tilt.shape[0] == order_range[1] - order_range[0]
    assert tilt.shape[1] == extracted.shape[1]

    assert isinstance(shear, np.ndarray)
    assert shear.ndim == 2
    assert shear.shape[0] == order_range[1] - order_range[0]
    assert shear.shape[1] == extracted.shape[1]



"""
Module that estimates the background scatter

Authors
-------
Ansgar Wehrhahn (ansgar.wehrhahn@physics.uu.se)

Version
-------
1.0 - 2d polynomial background scatter

License
-------
TODO

"""

import logging

import matplotlib.pyplot as plt
import numpy as np

from . import extract
from .util import polyfit2d, polyfit2d_2, make_index


def estimate_background_scatter(
    img,
    orders,
    column_range=None,
    extraction_width=0.1,
    scatter_degree=4,
    sigma_cutoff=2,
    border_width=10,
    plot=False,
    **kwargs
):
    """
    Estimate the background by fitting a 2d polynomial to interorder data

    Parameters
    ----------
    img : array[nrow, ncol]
        (flat) image data
    orders : array[nord, degree]
        order polynomial coefficients
    column_range : array[nord, 2], optional
        range of columns to use in each order (default: None == all columns)
    extraction_width : float, array[nord, 2], optional
        extraction width for each order, values below 1.5 are considered fractional, others as number of pixels (default: 0.1)
    scatter_degree : int, optional
        polynomial degree of the 2d fit for the background scatter (default: 4)
    plot : bool, optional
        wether to plot the fitted polynomial and the data or not (default: False)

    Returns
    -------
    array[nord+1, ncol]
        background scatter between orders
    array[nord+1, ncol]
        y positions of the interorder lines, the scatter values are taken from
    """

    if not isinstance(scatter_degree, (int, np.integer, tuple)):
        raise TypeError(
            "Expected integer value for scatter polynomial degree, got %s"
            % type(scatter_degree)
        )
    if isinstance(scatter_degree, tuple):
        if len(scatter_degree) != 2:
            raise ValueError(
                "Expected tuple of length 2, but got length %i" % len(scatter_degree)
            )
        types = [isinstance(i, (int, np.integer)) for i in scatter_degree]
        if not all(types):
            raise TypeError(
                "Expected integer value for scatter polynomial degree, got %s"
                % type(scatter_degree)
            )
        values = [i < 0 for i in scatter_degree]
        if any(values):
            raise ValueError(
                "Expected positive value for scatter polynomial degree, got %s"
                % str(scatter_degree)
            )
    elif scatter_degree < 0:
        raise ValueError(
            "Expected positive value for scatter polynomial degree, got %i"
            % scatter_degree
        )

    nrow, ncol = img.shape
    nord, _ = orders.shape

    if np.isscalar(extraction_width):
        extraction_width = np.tile([extraction_width, extraction_width], (nord, 1))
    else:
        extraction_width = np.asarray(extraction_width)
        if extraction_width.ndim == 1:
            extraction_width = np.tile(extraction_width, (nord, 1))
    if column_range is None:
        column_range = np.tile([0, ncol], (nord, 1))
    # Extend orders above and below orders
    orders = extract.extend_orders(orders, nrow)
    extraction_width = np.array(
        [extraction_width[0], *extraction_width, extraction_width[-1]]
    )
    column_range = np.array([column_range[0], *column_range, column_range[-1]])

    extraction_width = extract.fix_extraction_width(
        extraction_width, orders, column_range, ncol
    )

    # column_range = extract.fix_column_range(img, orders, extraction_width, column_range)
    column_range = column_range[1:-1]
    orders = orders[1:-1]
    extraction_width = extraction_width[1:-1]

    # Method 1: Select all pixels, but those known to be in orders
    bw = border_width
    mask = np.full(img.shape, True)
    if bw is not None and bw != 0:
        mask[:bw] = mask[-bw:] = mask[:, :bw] = mask[:, -bw:] = False
    for i in range(nord):
        left, right = column_range[i]
        left -= extraction_width[i, 1] * 2
        right += extraction_width[i, 0] * 2
        left = max(0, left)
        right = min(ncol, right)

        x_order = np.arange(left, right)
        y_order = np.polyval(orders[i], x_order)

        y_above = y_order + extraction_width[i, 1]
        y_below = y_order - extraction_width[i, 0]

        y_above = np.floor(y_above)
        y_below = np.ceil(y_below)

        index = make_index(y_below, y_above, left, right, zero=True)
        np.clip(index[0], 0, nrow - 1, out=index[0])

        mask[index] = False

    mask &= ~np.ma.getmask(img)

    y, x = np.indices(mask.shape)
    y, x = y[mask].ravel(), x[mask].ravel()
    z = np.ma.getdata(img[mask]).ravel()

    mask = z <= np.median(z) + sigma_cutoff * z.std()
    y, x, z = y[mask], x[mask], z[mask]

    coeff = polyfit2d(x, y, z, degree=scatter_degree, plot=plot)
    logging.debug("Background scatter coefficients: %s", str(coeff))

    if plot:
        # Calculate scatter at interorder positionsq
        yp, xp = np.indices(img.shape)
        back = np.polynomial.polynomial.polyval2d(xp, yp, coeff)

        plt.subplot(121)
        plt.title("Input Image + In-between Order traces")
        plt.xlabel("x [pixel]")
        plt.ylabel("y [pixel]")
        vmin, vmax = np.percentile(img - back, (5, 95))
        plt.imshow(img - back, vmin=vmin, vmax=vmax, aspect="equal", origin="lower")
        plt.plot(x, y, ",")

        plt.subplot(122)
        plt.title("2D fit to the scatter between orders")
        plt.xlabel("x [pixel]")
        plt.ylabel("y [pixel]")
        plt.imshow(back, vmin=0, vmax=abs(np.max(back)), aspect="equal", origin="lower")
        plt.show()

    return coeff



"""
Find the continuum level

Currently only splices orders together
First guess of the continuum is provided by the flat field
"""

from itertools import chain

import matplotlib.pyplot as plt
import numpy as np

from . import util

# np.seterr("raise")


def splice_orders(spec, wave, cont, sigm, scaling=True, plot=False):
    """
    Splice orders together so that they form a continous spectrum
    This is achieved by linearly combining the overlaping regions

    Parameters
    ----------
    spec : array[nord, ncol]
        Spectrum to splice, with seperate orders
    wave : array[nord, ncol]
        Wavelength solution for each point
    cont : array[nord, ncol]
        Continuum, blaze function will do fine as well
    sigm : array[nord, ncol]
        Errors on the spectrum
    scaling : bool, optional
        If true, the spectrum/continuum will be scaled to 1 (default: False)
    plot : bool, optional
        If true, will plot the spliced spectrum (default: False)

    Raises
    ------
    NotImplementedError
        If neighbouring orders dont overlap

    Returns
    -------
    spec, wave, cont, sigm : array[nord, ncol]
        spliced spectrum
    """
    nord, _ = spec.shape  # Number of sp. orders, Order length in pixels

    # Just to be extra safe that they are all the same
    mask = np.ma.getmaskarray(spec) | (spec == 0) | (cont == 0)
    spec = np.ma.masked_array(spec, mask=mask)
    wave = np.ma.masked_array(np.ma.getdata(wave), mask=mask)
    cont = np.ma.masked_array(np.ma.getdata(cont), mask=mask)
    sigm = np.ma.masked_array(np.ma.getdata(sigm), mask=mask)

    if scaling:
        # Scale everything to roughly the same size, around spec/blaze = 1
        scale = np.ma.median(spec / cont, axis=1)
        cont *= scale[:, None]

    if plot:
        plt.subplot(411)
        plt.title("Before")
        for i in range(spec.shape[0]):
            plt.plot(wave[i], spec[i] / cont[i])
        plt.ylim([0, 2])

        plt.subplot(412)
        plt.title("Before Error")
        for i in range(spec.shape[0]):
            plt.plot(wave[i], sigm[i] / cont[i])
        plt.ylim((0, np.ma.median(sigm[i] / cont[i]) * 2))

    # Order with largest signal, everything is scaled relative to this order
    iord0 = np.argmax(np.ma.median(spec / cont, axis=1))

    # Loop from iord0 outwards, first to the top, then to the bottom
    tmp0 = chain(range(iord0, 0, -1), range(iord0, nord - 1))
    tmp1 = chain(range(iord0 - 1, -1, -1), range(iord0 + 1, nord))

    for iord0, iord1 in zip(tmp0, tmp1):
        # Get data for current order
        # Note that those are just references to parts of the original data
        # any changes will also affect spec, wave, cont, and sigm
        s0, s1 = spec[iord0], spec[iord1]
        w0, w1 = wave[iord0], wave[iord1]
        c0, c1 = cont[iord0], cont[iord1]
        u0, u1 = sigm[iord0], sigm[iord1]

        # Calculate Overlap
        i0 = np.ma.where((w0 >= np.ma.min(w1)) & (w0 <= np.ma.max(w1)))
        i1 = np.ma.where((w1 >= np.ma.min(w0)) & (w1 <= np.ma.max(w0)))

        # Orders overlap
        if i0[0].size > 0 and i1[0].size > 0:
            # Interpolate the overlapping region onto the wavelength grid of the other order
            tmpS0 = util.bezier_interp(w1, s1, w0[i0])
            tmpB0 = util.bezier_interp(w1, c1, w0[i0])
            tmpU0 = util.bezier_interp(w1, u1, w0[i0])

            tmpS1 = util.bezier_interp(w0, s0, w1[i1])
            tmpB1 = util.bezier_interp(w0, c0, w1[i1])
            tmpU1 = util.bezier_interp(w0, u0, w1[i1])

            # Combine the two orders weighted by the relative error
            wgt0 = np.ma.vstack([c0[i0].data / u0[i0].data, tmpB0 / tmpU0]) ** 2
            wgt1 = np.ma.vstack([c1[i1].data / u1[i1].data, tmpB1 / tmpU1]) ** 2

            s0[i0], utmp = np.ma.average(
                np.ma.vstack([s0[i0], tmpS0]), axis=0, weights=wgt0, returned=True
            )
            c0[i0] = np.ma.average([c0[i0], tmpB0], axis=0, weights=wgt0)
            u0[i0] = c0[i0] * utmp ** -0.5

            s1[i1], utmp = np.ma.average(
                np.ma.vstack([s1[i1], tmpS1]), axis=0, weights=wgt1, returned=True
            )
            c1[i1] = np.ma.average([c1[i1], tmpB1], axis=0, weights=wgt1)
            u1[i1] = c1[i1] * utmp ** -0.5
        else:  # Orders dont overlap
            continue
            raise NotImplementedError("Orders don't overlap, please test")
            c0 *= util.top(s0 / c0, 1, poly=True)
            scale0 = util.top(s0 / c0, 1, poly=True)
            scale0 = np.polyfit(w0, scale0, 1)

            scale1 = util.top(s1 / c1, 1, poly=True)
            scale1 = np.polyfit(w1, scale1, 1)

            xx = np.linspace(np.min(w0), np.max(w1), 100)

            # TODO test this
            # scale = np.sum(scale0[0] * scale1[0] * xx * xx + scale0[0] * scale1[1] * xx + scale1[0] * scale0[1] * xx + scale1[1] * scale0[1])
            scale = scale0[::-1, None] * scale1[None, ::-1]
            scale = np.sum(np.polynomial.polynomial.polyval2d(xx, xx, scale)) / np.sum(
                np.polyval(scale1, xx) ** 2
            )
            s1 *= scale

    if plot:
        plt.subplot(413)
        plt.title("After")
        for i in range(nord):
            plt.plot(wave[i], spec[i] / cont[i], label="order=%i" % i)
        plt.ylim((0, 2))

        plt.subplot(414)
        plt.title("Error")
        for i in range(nord):
            plt.plot(wave[i], sigm[i] / cont[i], label="order=%i" % i)
        plt.ylim((0, np.ma.median(sigm[i] / cont[i]) * 2))
        plt.show()

    return spec, wave, cont, sigm


class Plot_Normalization:
    def __init__(self, wsort, sB, new_wave, contB, iteration=0):
        plt.ion()
        self.fig = plt.figure()
        self.fig.suptitle(f"Iteration: {iteration}")

        self.ax = self.fig.add_subplot(111)
        self.line1 = self.ax.plot(wsort, sB, label="Spectrum")[0]
        self.line2 = self.ax.plot(new_wave, contB, label="Continuum Fit")[0]
        plt.legend()

        plt.show()

    def plot(self, wsort, sB, new_wave, contB, iteration):
        self.fig.suptitle(f"Iteration: {iteration}")
        self.line1.set_xdata(wsort)
        self.line1.set_ydata(sB)
        self.line2.set_xdata(new_wave)
        self.line2.set_ydata(contB)

        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def close(self):
        plt.ioff()
        plt.close()


def continuum_normalize(
    spec,
    wave,
    cont,
    sigm,
    iterations=10,
    smooth_initial=1e5,
    smooth_final=5e6,
    scale_vert=1,
    plot=True,
):
    """ Fit a continuum to a spectrum by slowly approaching it from the top.
    We exploit here that the continuum varies only on large wavelength scales, while individual lines act on much smaller scales

    TODO automatically find good parameters for smooth_initial and smooth_final
    TODO give variables better names

    Parameters
    ----------
    spec : masked array of shape (nord, ncol)
        Observed input spectrum, masked values describe column ranges
    wave : masked array of shape (nord, ncol)
        Wavelength solution of the spectrum
    cont : masked array of shape (nord, ncol)
        Initial continuum guess, for example based on the blaze
    sigm : masked array of shape (nord, ncol)
        Uncertainties of the spectrum
    iterations : int, optional
        Number of iterations of the algorithm, 
        note that runtime roughly scales with the number of iterations squared 
        (default: 10)
    smooth_initial : float, optional
        Smoothing parameter in the initial runs, usually smaller than smooth_final (default: 1e5)
    smooth_final : float, optional
        Smoothing parameter of the final run (default: 5e6)
    scale_vert : float, optional
        Vertical scale of the spectrum. Usually 1 if a previous normalization exists (default: 1)
    plot : bool, optional
        Wether to plot the current status and results or not (default: True)

    Returns
    -------
    cont : masked array of shape (nord, ncol)
        New continuum
    """

    nord, ncol = spec.shape

    par2 = 1e-4
    par4 = 0.01 * (1 - np.clip(2, None, 1 / np.sqrt(np.ma.median(spec))))

    b = np.clip(cont, 1, None)
    for i in range(nord):
        b[i, ~b.mask[i]] = util.middle(b[i, ~b.mask[i]], 1)
    cont = b

    # Create new equispaced wavelength grid
    wmin = np.ma.min(wave)
    wmax = np.ma.max(wave)
    dwave = np.abs(wave[nord // 2, ncol // 2] - wave[nord // 2, ncol // 2 - 1]) * 0.5
    nwave = np.ceil((wmax - wmin) / dwave) + 1
    new_wave = np.linspace(wmin, wmax, nwave, endpoint=True)

    # Combine all orders into one big spectrum, sorted by wavelength
    wsort, j, index = np.unique(
        wave.compressed(), return_index=True, return_inverse=True
    )
    sB = (spec / cont).compressed()[j]

    # Get initial weights for each point
    weight = util.middle(sB, 0.5, x=wsort - wmin)
    weight = weight / util.middle(weight, 3 * smooth_initial) + np.concatenate(
        ([0], 2 * weight[1:-1] - weight[0:-2] - weight[2:], [0])
    )
    weight = np.clip(weight, 0, None)
    # TODO for some reason the interpolation messes up, use linear instead for now
    # weight = util.safe_interpolation(wsort, weight, new_wave)
    weight = np.interp(new_wave, wsort, weight)
    weight /= np.max(weight)

    # Interpolate Spectrum onto the new grid
    # ssB = util.safe_interpolation(wsort, sB, new_wave)
    ssB = np.interp(new_wave, wsort, sB)
    # Keep the scale of the continuum
    bbb = util.middle(cont.compressed()[j], 1)

    contB = 1
    for i in range(iterations):
        # Find new approximation of the top, smoothed by some parameter
        c = ssB / contB
        for _ in range(iterations):
            _c = util.top(
                c, smooth_initial, eps=par2, weight=weight, lambda2=smooth_final
            )
            c = np.clip(_c, c, None)
        c = (
            util.top(c, smooth_initial, eps=par4, weight=weight, lambda2=smooth_final)
            * contB
        )

        # Scale it and update the weights of each point
        contB = c * scale_vert
        contB = util.middle(contB, 1)
        weight = np.clip(ssB / contB, None, contB / np.clip(ssB, 1, None))

        # Plot the intermediate results
        if plot:
            if i == 0:
                p = Plot_Normalization(wsort, sB, new_wave, contB, i)
            else:
                p.plot(wsort, sB, new_wave, contB, i)

    # Need to close the plot afterwards
    if plot:
        p.close()

    # Calculate the new continuum from intermediate values
    # new_cont = util.safe_interpolation(new_wave, contB, wsort)
    new_cont = np.interp(wsort, new_wave, contB)
    cont[~cont.mask] = (new_cont * bbb)[index]

    # Final output plot
    if plot:
        plt.plot(wave.ravel(), spec.ravel(), label="spec")
        plt.plot(wave.ravel(), cont.ravel(), label="cont")
        plt.legend(loc="best")
        plt.xlabel("Wavelength [A]")
        plt.ylabel("Flux")
        plt.show()

    return cont



import logging

import numpy as np
import matplotlib.pyplot as plt

from .util import gaussfit4 as gaussfit, make_index, gaussval2 as gaussval
from .extract import fix_column_range


def estimate_extraction_width(img, orders, column_range, plot=False):
    nrow, ncol = img.shape
    nord, _ = orders.shape
    extraction_width = np.zeros((nord, 2), dtype=int)

    for i in range(nord):
        # first guess, half way to the next order
        # To order above
        if i < nord - 1:
            beg = max(column_range[[i, i + 1], 0])
            end = min(column_range[[i, i + 1], 1])
            x = np.arange(beg, end)
            y = np.polyval(orders[i], x)
            y_above = np.polyval(orders[i + 1], x)
            width_above = int(np.mean(y_above - y) // 2)

        # To order below
        if i > 0:
            beg = max(column_range[[i - 1, i], 0])
            end = min(column_range[[i - 1, i], 1])
            x = np.arange(beg, end)
            y = np.polyval(orders[i], x)
            y_below = np.polyval(orders[i - 1], x)
            width_below = int(np.mean(y - y_below) // 2)
        else:
            width_below = width_above

        if i == nord - 1:
            width_above = width_below

        beg, end = column_range[i]
        x = np.arange(beg, end)
        y = np.polyval(orders[i], x)

        y_int = y.astype(int)
        y_above = y_int + width_above
        y_below = y_int - width_below

        if np.ma.any(y_above >= nrow) or np.ma.any(y_below < 0):
            beg, end = np.where((y_above < nrow) & (y_below >= 0))[0][[0, -1]]

        index = make_index(y_int - width_below, y_int + width_above, beg, end, True)

        slitf = np.ma.sum(img[index], axis=1)

        # TODO fit rectangular profile

        # width = int(4 * coef[2])

        # plt.plot(p, slitf)
        # plt.plot(p, gaussval(p, *coef))
        # plt.vlines([coef[1] - width, coef[1] + width], slitf.min(), slitf.max())
        # plt.show()

        # width_below = min(width_below, width)
        # width_above = min(width_above, width)

        # index = make_index(y_int - width_below, y_int + width_above, beg, end, True)

        # plt.imshow(np.log(img[index]), aspect="auto", origin="lower")
        # plt.show()

        extraction_width[i] = [width, width]

    return extraction_width



import pytest
import numpy as np

from pyreduce.extract import extract


def test_normflat(flat, orders, settings, order_range, scatter):
    flat, fhead = flat
    orders, column_range = orders
    settings = settings["norm_flat"]

    norm, _, blaze, _ = extract(
        flat,
        orders,
        scatter=scatter,
        gain=fhead["e_gain"],
        readnoise=fhead["e_readn"],
        dark=fhead["e_drk"],
        column_range=column_range,
        order_range=order_range,
        extraction_type="normalize",
        extraction_width=settings["extraction_width"],
        threshold=settings["threshold"],
        lambda_sf=settings["smooth_slitfunction"],
        lambda_sp=settings["smooth_spectrum"],
        swath_width=settings["swath_width"],
        plot=False,
    )

    assert isinstance(norm, np.ndarray)
    assert norm.ndim == flat.ndim
    assert norm.shape[0] == flat.shape[0]
    assert norm.shape[1] == flat.shape[1]
    assert norm.dtype == flat.dtype
    assert np.ma.min(norm) > 0
    assert not np.any(np.isnan(norm))

    assert isinstance(blaze, np.ndarray)
    assert blaze.ndim == 2
    assert blaze.shape[0] == order_range[1] - order_range[0]
    assert blaze.shape[1] == flat.shape[1]
    assert np.issubdtype(blaze.dtype, np.floating)
    assert not np.any(np.isnan(blaze))

    for i, j in enumerate(range(order_range[0], order_range[1])):
        cr = column_range[j]
        assert np.all(blaze[i, : cr[0]].mask == True)
        assert np.all(blaze[i, cr[1] :].mask == True)



"""
Simple usage example for PyReduce
Loads a sample UVES dataset, and runs the full extraction
"""

import os.path
import pyreduce
from pyreduce import datasets


# define parameters
instrument = "UVES"
target = "HD132205"
night = "2010-04-01"
mode = "middle"
steps = (
    "bias",
    "flat",
    "orders",
    "norm_flat",
    "wavecal",
    "curvature",
    "science",
    "continuum",
    "finalize",
)

# some basic settings
# Expected Folder Structure: base_dir/datasets/HD132205/*.fits.gz
# Feel free to change this to your own preference, values in curly brackets will be replaced with the actual values {}

# load dataset (and save the location)
base_dir = datasets.UVES_HD132205()
input_dir = "{target}/"
output_dir = "reduced/{instrument}/{target}/{night}/{mode}"

config = pyreduce.configuration.get_configuration_for_instrument(instrument, plot=True)

pyreduce.reduce.main(
    instrument,
    target,
    night,
    mode,
    steps,
    base_dir=base_dir,
    input_dir=input_dir,
    output_dir=output_dir,
    configuration=config,
    order_range=(1, 21),
)



import pytest
import numpy as np

import os
from os.path import join

from pyreduce import util


@pytest.fixture
def mask_dir():
    mask_dir = os.path.dirname(__file__)
    mask_dir = os.path.join(mask_dir, "../pyreduce", "masks")
    return mask_dir


def test_load_mask(instrument, mode, mask_dir):
    mask_file = join(mask_dir, "mask_%s_%s.fits.gz" % (instrument.lower(), mode))
    mask, _ = util.load_fits(mask_file, instrument, mode, extension=0)
    mask = ~mask.data.astype(bool)  # REDUCE mask are inverse to numpy masks

    assert isinstance(mask, np.ndarray)
    assert not np.all(mask)



"""
Calculate the tilt based on a reference spectrum with high SNR, e.g. Wavelength calibration image

Authors
-------
Nikolai Piskunov
Ansgar Wehrhahn

Version
--------
0.9 - NP - IDL Version
1.0 - AW - Python Version

License
-------
....
"""

import logging
import warnings

import matplotlib.pyplot as plt
import numpy as np
from scipy import signal

from numpy.polynomial.polynomial import polyval2d
from skimage.filters import threshold_otsu

from .extract import fix_extraction_width, extend_orders, fix_column_range
from .util import make_index, gaussfit4 as gaussfit, polyfit2d


class ProgressPlot:
    def __init__(self, ncol, width, height):
        plt.ion()

        fig, (ax1, ax2) = plt.subplots(ncols=2)

        fig.suptitle("Curvature in each order")

        line1, = ax1.plot(np.arange(ncol) + 1)
        line2, = ax1.plot(0, 0, "d")
        ax1.set_yscale("log")

        lines = [None] * height
        for i in range(height):
            lines[i], = ax2.plot(
                np.arange(-width, width + 1), np.arange(-width, width + 1)
            )

        line3, = ax2.plot(np.arange(height), "r--")
        line4, = ax2.plot(np.arange(height), "rx")
        ax2.set_xlim((-width, width))
        ax2.set_ylim((0, height + 5))

        self.ncol = ncol
        self.width = width * 2 + 1
        self.height = height

        self.fig = fig
        self.ax1 = ax1
        self.ax2 = ax2
        self.line1 = line1
        self.line2 = line2
        self.line3 = line3
        self.line4 = line4
        self.lines = lines

    def update_plot1(self, vector, peak, offset=0):
        data = np.ones(self.ncol)
        data[offset : len(vector) + offset] = np.clip(vector, 1, None)
        self.line1.set_ydata(data)
        self.line2.set_xdata(peak)
        self.line2.set_ydata(data[peak])
        self.ax1.set_ylim((data.min(), data.max()))
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def update_plot2(self, segments, tilt, shear, positions, values):
        l4_y = np.full(self.height, np.nan)
        for i, (s, v) in enumerate(zip(segments, values)):
            s, v = s - s.min(), v - s.min()
            s, v = s / s.max() * 5, v / s.max() * 5
            l4_y[i] = v + i
            self.lines[i].set_ydata(s + i)
        for i in range(len(segments), self.height):
            self.lines[i].set_ydata(np.full(self.width, np.nan))

        y = np.arange(0, self.height) - self.height // 2
        x = np.polyval((shear, tilt, 0), y)
        y += np.arange(self.height)
        y += self.height // 2
        self.line3.set_xdata(x)
        self.line3.set_ydata(y)

        if positions.size > self.height:
            positions = positions[:self.height]
        elif positions.size < self.height:
            positions = np.concatenate((positions, np.zeros(self.height - positions.size)))

        self.line4.set_xdata(positions)
        self.line4.set_ydata(l4_y)

        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def close(self):
        plt.close()
        plt.ioff()


class Curvature:
    def __init__(
        self,
        orders,
        extraction_width=0.5,
        column_range=None,
        order_range=None,
        width=9,
        threshold=10,
        fit_degree=2,
        sigma_cutoff=3,
        max_iter=None,
        mode="1D",
        plot=False,
        verbose=0,
        curv_degree=2,
    ):
        self.orders = orders
        self.extraction_width = extraction_width
        self.column_range = column_range
        if order_range is None:
            order_range = (0, self.nord)
        self.order_range = order_range
        self.width = width
        self.threshold = threshold
        self.fit_degree = fit_degree
        self.sigma_cutoff = sigma_cutoff
        if max_iter is None:
            max_iter = np.inf
        self.max_iter = max_iter
        self.mode = mode
        self.plot = plot
        self.verbose = verbose
        self.curv_degree = curv_degree

    @property
    def nord(self):
        return self.orders.shape[0]

    @property
    def n(self):
        return self.order_range[1] - self.order_range[0]

    def _fix_inputs(self, original):
        nrow, ncol = original.shape

        orders = self.orders
        extraction_width = self.extraction_width
        column_range = self.column_range

        if np.isscalar(extraction_width):
            extraction_width = np.tile([extraction_width], [self.nord, 2])
        if column_range is None:
            column_range = np.tile([0, ncol], [self.nord, 1])

        orders = extend_orders(orders, nrow)
        extraction_width = np.array(
            [extraction_width[0], *extraction_width, extraction_width[-1]]
        )
        column_range = np.array([column_range[0], *column_range, column_range[-1]])

        # Fix column range, so that all extractions are fully within the image
        extraction_width = fix_extraction_width(
            extraction_width, orders, column_range, ncol
        )
        column_range = fix_column_range(
            original, orders, extraction_width, column_range
        )

        extraction_width = extraction_width[1:-1]
        column_range = column_range[1:-1]
        orders = orders[1:-1]

        self.column_range = column_range[self.order_range[0] : self.order_range[1]]
        self.extraction_width = extraction_width[
            self.order_range[0] : self.order_range[1]
        ]
        self.orders = orders[self.order_range[0] : self.order_range[1]]
        self.order_range = (0, self.n)

    def _find_peaks(self, vec, cr):
        # This should probably be the same as in the wavelength calibration
        vec -= np.ma.min(vec)
        vec = np.ma.filled(vec, 0)
        height = np.quantile(vec, 0.1) * self.threshold
        peaks, _ = signal.find_peaks(vec, prominence=height)

        # Remove peaks at the edge
        peaks = peaks[(peaks >= self.width + 1) & (peaks < len(vec) - self.width - 1)]
        # Remove the offset, due to vec being a subset of extracted
        peaks += cr[0]
        return vec, peaks

    def _determine_curvature_single_line(self, original, peak, ycen, xwd):
        nrow, ncol = original.shape
        height = np.sum(xwd) + 1

        # look at +- width pixels around the line
        #:array of shape (2*width + 1,): indices of the pixels to the left and right of the line peak
        index_x = np.arange(-self.width, self.width + 1)
        #:array of shape (height,): stores the peak positions of the fits to each row
        xcen = np.zeros(height)
        vcen = np.zeros(height)
        wcen = np.zeros(height)
        #:array of shape (height,): indices of the rows in the order, with 0 being the central row
        xind = np.arange(-xwd[0], xwd[1] + 1)
        #:array of shape (height,): Scatter of the values within the row, to seperate in order and out of order rows
        deviation = np.zeros(height)

        segments = []

        # Extract short horizontal strip for each row in extraction width
        # Then fit a gaussian to each row, to find the center of the line
        x = peak + index_x
        x = x[(x >= 0) & (x < ncol)]
        xmin, xmax = x[0], x[-1] + 1
        x = np.ma.masked_array(x)
        for i, irow in enumerate(xind):
            # Trying to access values outside the image
            assert not np.any((ycen + irow)[xmin : xmax] >= nrow)

            # Just cutout this one row
            idx = make_index(ycen + irow, ycen + irow, xmin, xmax)
            segment = original[idx][0]
            segments += [segment]

            if np.all(np.ma.getmask(segment)):
                # If this row is masked, this will happen
                # It will be ignored by the thresholding anyway
                xcen[i] = np.mean(x)
                deviation[i] = 0
            else:
                try:
                    x.mask = segment.mask
                    coef = gaussfit(x, segment)
                    # Store line center
                    xcen[i] = coef[1]
                    wcen[i] = coef[2]
                    vcen[i] = coef[0] + coef[3]
                    # Store the variation within the row
                    deviation[i] = np.ma.std(segment)
                except RuntimeError:
                    xcen[i] = np.ma.mean(x)
                    deviation[i] = 0
                except TypeError:
                    xcen[i] = np.ma.mean(x)
                    deviation[i] = 0
        # Seperate in order pixels from out of order pixels
        # TODO: actually we want to weight them by the slitfunction?
        if not np.all(deviation == 0):
            idx = deviation > threshold_otsu(deviation)
        else:
            idx = np.full(deviation.shape, True)

        # Linear fit to slit image
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                w = np.sqrt(1/wcen[idx])
                coef = np.polyfit(xind[idx], xcen[idx], self.curv_degree, w=w)
        except:
            logging.warning("Could not fit curvature to line, using 0 instead.")
            coef = [0] * self.curv_degree

        if self.curv_degree == 1:
            tilt, shear = coef[0], 0
        elif self.curv_degree == 2:
            tilt, shear = coef[1], coef[0]
        else:
            raise ValueError("Only curvature degrees 1 and 2 are supported")

        if self.plot and self.verbose >= 2:
            self.progress.update_plot2(segments, tilt, shear, xcen - peak, vcen)
        return tilt, shear

    def _fit_curvature_single_order(self, peaks, tilt, shear):

        # Make them masked arrays to avoid copying the data all the time
        # Updating the mask updates all of them (as it is not copied)
        mask = np.full(peaks.shape, False)
        peaks = np.ma.masked_array(peaks, mask=mask)
        tilt = np.ma.masked_array(tilt, mask=mask)
        shear = np.ma.masked_array(shear, mask=mask)

        if len(peaks) <= self.fit_degree:
            coef_tilt = np.ma.polyfit(peaks, tilt, len(peaks) - 1)
            coef_shear = np.ma.polyfit(peaks, shear, len(peaks) - 1)
        else:
            # Fit a 2nd order polynomial through all individual lines
            # And discard obvious outliers
            iteration = 0
            while iteration < self.max_iter:
                iteration += 1
                coef_tilt = np.ma.polyfit(peaks, tilt, self.fit_degree)
                coef_shear = np.ma.polyfit(peaks, shear, self.fit_degree)

                diff = np.polyval(coef_tilt, peaks) - tilt
                idx1 = np.ma.abs(diff) >= np.ma.std(diff) * self.sigma_cutoff
                mask |= idx1

                diff = np.polyval(coef_shear, peaks) - shear
                idx2 = np.ma.abs(diff) >= np.ma.std(diff) * self.sigma_cutoff
                mask |= idx2

                # if no maximum iteration is given, go on forever
                if np.ma.all(~idx1) and np.ma.all(~idx2):
                    break
                if np.all(mask):
                    logging.error("Could not fit polynomial to the data")
                    mask[:] = False
                    break

        coef_tilt = np.ma.polyfit(peaks, tilt, self.fit_degree)
        coef_shear = np.ma.polyfit(peaks, shear, self.fit_degree)

        return coef_tilt, coef_shear, peaks

    def _determine_curvature_all_lines(self, original, extracted):
        ncol = original.shape[1]
        # Store data from all orders
        all_peaks = []
        all_tilt = []
        all_shear = []
        plot_vec = []

        for j in range(self.n):
            if self.n < 10 or j % 5 == 0:
                logging.info("Calculating tilt of order %i out of %i", j + 1, self.n)
            else:
                logging.debug("Calculating tilt of order %i out of %i", j + 1, self.n)

            cr = self.column_range[j]
            xwd = self.extraction_width[j]
            ycen = np.polyval(self.orders[j], np.arange(ncol)).astype(int)

            # Find peaks
            vec = extracted[j, cr[0] : cr[1]]
            vec, peaks = self._find_peaks(vec, cr)
            npeaks = len(peaks)

            # Determine curvature for each line seperately
            tilt = np.zeros(npeaks)
            shear = np.zeros(npeaks)
            for ipeak, peak in enumerate(peaks):
                if self.plot and self.verbose >= 2:
                    self.progress.update_plot1(vec, peak, cr[0])
                tilt[ipeak], shear[ipeak] = self._determine_curvature_single_line(
                    original, peak, ycen, xwd
                )

            # Store results
            all_peaks += [peaks]
            all_tilt += [tilt]
            all_shear += [shear]
            plot_vec += [vec]
        return all_peaks, all_tilt, all_shear, plot_vec

    def fit(self, peaks, tilt, shear):
        if self.mode == "1D":
            coef_tilt = np.zeros((self.n, self.fit_degree + 1))
            coef_shear = np.zeros((self.n, self.fit_degree + 1))
            for j in range(self.n):
                coef_tilt[j], coef_shear[j], _ = self._fit_curvature_single_order(
                    peaks[j], tilt[j], shear[j]
                )
        elif self.mode == "2D":
            x = np.concatenate(peaks)
            y = [np.full(len(p), i) for i, p in enumerate(peaks)]
            y = np.concatenate(y)
            z = np.concatenate(tilt)
            coef_tilt = polyfit2d(x, y, z, degree=self.fit_degree)
            z = np.concatenate(shear)
            coef_shear = polyfit2d(x, y, z, degree=self.fit_degree)
        else:
            raise ValueError(
                f"Value for 'mode' not understood. Expected one of ['1D', '2D'] but got {self.mode}"
            )

        return coef_tilt, coef_shear

    def eval(self, peaks, order, coef_tilt, coef_shear):
        if self.mode == "1D":
            tilt = np.zeros(peaks.shape)
            shear = np.zeros(peaks.shape)
            for i in np.unique(order):
                idx = order == i
                tilt[idx] = np.polyval(coef_tilt[i], peaks[idx])
                shear[idx] = np.polyval(coef_shear[i], peaks[idx])
        elif self.mode == "2D":
            tilt = polyval2d(peaks, order, coef_tilt)
            shear = polyval2d(peaks, order, coef_shear)
        else:
            raise ValueError(
                f"Value for 'mode' not understood. Expected one of ['1D', '2D'] but got {self.mode}"
            )
        return tilt, shear

    def plot_results(
        self, ncol, plot_peaks, plot_vec, plot_tilt, plot_shear, tilt_x, shear_x
    ):
        fig, axes = plt.subplots(nrows=self.n // 2 + self.n % 2, ncols=2, squeeze=False)
        fig.suptitle("Peaks")
        fig1, axes1 = plt.subplots(
            nrows=self.n // 2 + self.n % 2, ncols=2, squeeze=False
        )
        fig1.suptitle("1st Order Curvature")
        fig2, axes2 = plt.subplots(
            nrows=self.n // 2 + self.n % 2, ncols=2, squeeze=False
        )
        fig2.suptitle("2nd Order Curvature")
        plt.subplots_adjust(hspace=0)

        for j in range(self.n):
            cr = self.column_range[j]
            peaks = plot_peaks[j]
            vec = np.clip(plot_vec[j], 0, None)
            tilt = plot_tilt[j]
            shear = plot_shear[j]
            x = np.arange(cr[0], cr[1])

            order = np.full(len(x), j)
            t, s = self.eval(x, order, tilt_x, shear_x)

            # Figure Peaks found (and used)
            axes[j // 2, j % 2].plot(np.arange(cr[0], cr[1]), vec)
            axes[j // 2, j % 2].plot(peaks, vec[peaks - cr[0]], "+")
            axes[j // 2, j % 2].set_xlim([0, ncol])
            axes[j // 2, j % 2].set_yscale("log")
            if j not in (self.n - 1, self.n - 2):
                axes[j // 2, j % 2].get_xaxis().set_ticks([])

            # Figure 1st order
            axes1[j // 2, j % 2].plot(peaks, tilt, "rx")
            axes1[j // 2, j % 2].plot(x, t)
            axes1[j // 2, j % 2].set_xlim(0, ncol)

            lower = t.min() * (0.5 if t.min() > 0 else 1.5)
            upper = t.max() * (1.5 if t.max() > 0 else 0.5)
            axes1[j // 2, j % 2].set_ylim(lower, upper)
            if j not in (self.n - 1, self.n - 2):
                axes1[j // 2, j % 2].get_xaxis().set_ticks([])

            # Figure 2nd order
            axes2[j // 2, j % 2].plot(peaks, shear, "rx")
            axes2[j // 2, j % 2].plot(x, s)
            axes2[j // 2, j % 2].set_xlim(0, ncol)

            lower = s.min() * (0.5 if s.min() > 0 else 1.5)
            upper = s.max() * (1.5 if s.max() > 0 else 0.5)
            axes2[j // 2, j % 2].set_ylim(lower, upper)
            if j not in (self.n - 1, self.n - 2):
                axes2[j // 2, j % 2].get_xaxis().set_ticks([])

        plt.show()

    def plot_comparison(self, original, tilt, shear, peaks):
        _, ncol = original.shape
        output = np.zeros((np.sum(self.extraction_width) + self.nord, ncol))
        pos = [0]
        x = np.arange(ncol)
        for i in range(self.nord):
            ycen = np.polyval(self.orders[i], x)
            yb = ycen - self.extraction_width[i, 0]
            yt = ycen + self.extraction_width[i, 1]
            xl, xr = self.column_range[i]
            index = make_index(yb, yt, xl, xr)
            yl = pos[i]
            yr = pos[i] + index[0].shape[0]
            output[yl:yr, xl:xr] = original[index]
            pos += [yr]

        vmin, vmax = np.percentile(output[output != 0], (5, 95))
        plt.imshow(output, vmin=vmin, vmax=vmax, origin="lower", aspect="auto")
        
        for i in range(self.nord):
            for p in peaks[i]:
                ew = self.extraction_width[i]
                x = np.zeros(ew[0] + ew[1] + 1)
                y = np.arange(-ew[0], ew[1] + 1)
                for j, yt in enumerate(y):
                    x[j] = p + yt * tilt[i, p] + yt**2 * shear[i, p]
                y += pos[i] + ew[0]
                plt.plot(x, y, "r")
        
        locs = np.sum(self.extraction_width, axis=1) + 1
        locs = [0, *np.cumsum(locs)[:-1]]
        plt.yticks(locs, range(len(locs)))
        plt.xlabel("x [pixel]")
        plt.ylabel("order")
        plt.show()

    def execute(self, extracted, original):
        _, ncol = original.shape

        self._fix_inputs(original)

        if self.plot and self.verbose >= 2:
            height = np.sum(self.extraction_width, axis=1).max() + 1
            self.progress = ProgressPlot(ncol, self.width, height)

        peaks, tilt, shear, vec = self._determine_curvature_all_lines(
            original, extracted
        )
        coef_tilt, coef_shear = self.fit(peaks, tilt, shear)

        if self.plot:
            self.plot_results(ncol, peaks, vec, tilt, shear, coef_tilt, coef_shear)

        iorder, ipeaks = np.indices(extracted.shape)
        tilt, shear = self.eval(ipeaks, iorder, coef_tilt, coef_shear)

        if self.plot:
            self.plot_comparison(original, tilt, shear, peaks)


        return tilt, shear



# Define Version
from ._version import get_versions
__version__ = get_versions()["version"]
del get_versions

# Start logging
from . import util
util.start_logging(None)

# Make sure that clibraries exists and are compilled
import os
dirname = os.path.dirname(__file__)
fname = os.path.join(dirname, "clib", "_slitfunc_2d.c")

if not os.path.exists(fname):
    from .clib import build_extract
    build_extract.build()
    del build_extract
del os

# Load externally available modules
from . import reduce, datasets, instruments, util, configuration



import os.path
import numpy as np
import matplotlib.pyplot as plt
import pyreduce
from pyreduce import datasets, instruments, util
from pyreduce.wavelength_calibration import (
    WavelengthCalibration as WavelengthCalibrationModule
)


def func_wavecal(
    degree, thar, instrument, mode, threshold, iterations, wavecal_mode, shift_window
):
    reference = instruments.instrument_info.get_wavecal_filename(None, instrument, mode)
    reference = np.load(reference, allow_pickle=True)
    linelist = reference["cs_lines"]

    module = WavelengthCalibrationModule(
        plot=False,
        manual=False,
        degree=degree,
        threshold=threshold,
        iterations=iterations,
        mode=wavecal_mode,
        shift_window=shift_window,
    )
    wave, coef = module.execute(thar, linelist)

    return module.aic

def func_freq_comb(degree, comb, wave, threshold, wavecal_mode):
    module = WavelengthCalibrationModule(
            plot=False,
            degree=degree,
            threshold=threshold,
            mode=wavecal_mode,
            lfc_peak_width=3,
            nstep=8
        )
    wave = module.frequency_comb(comb, wave)
    return module.aic

# define parameters
instrument = "HARPS"
target = "HD109200"
night = "2015-04-09"
mode = "red"

# some basic settings
# Expected Folder Structure: base_dir/datasets/HD132205/*.fits.gz
# Feel free to change this to your own preference, values in curly brackets will be replaced with the actual values {}

# load dataset (and save the location)
base_dir = "/DATA/PyReduce/"
input_dir = "FrequencyComb/raw"
output_dir = f"FrequencyComb/reduced_{mode}"

f = os.path.join(base_dir, output_dir, "harps_red.thar.npz")
data = np.load(f, allow_pickle=True)
thar = data["thar"]
wave = data["wave"]

f = os.path.join(base_dir, output_dir, "harps_red.comb.npz")
data = np.load(f, allow_pickle=True)
comb = data["comb"]

ndim = 2
kwargs = {
    "thar": thar,
    "instrument": instrument,
    "mode": mode,
    "threshold": 100,
    "iterations": 3,
    "wavecal_mode": f"{ndim}D",
    "shift_window": 0.01,
}
kwargs_comb = {"wave": wave, "comb": comb, "threshold":100, "wavecal_mode": f"{ndim}D"}

shape = tuple([20] * ndim)
grid = np.zeros((*shape, ndim), int)
for i in np.ndindex(shape):
    grid[i] = i

# matrix = util.gridsearch(func_wavecal, grid, kwargs=kwargs)
matrix = util.gridsearch(func_freq_comb, grid, kwargs=kwargs_comb)

np.save(f"matrix_comb_{ndim}D.npy", matrix)

if ndim == 1:
    plt.plot(grid[:, 0], matrix)
elif ndim == 2:
    plt.imshow(matrix)
    # plt.xticks(labels=grid[0])
    # plt.yticks(labels=grid[:, 0])
plt.show()



import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from pyreduce import cwrappers, extract

input_dir = "./datasets/reduced/UVES/HD132205/2010-04-02/middle/"
raw_dir = "./datasets/HD132205/"

hdu = fits.open(input_dir + "uves_middle.flat.fits")
img = hdu[0].data
nrow, ncol = img.shape

data = np.load(input_dir + "uves_middle.ord_default.npz", allow_pickle=True)
orders = data["orders"]
column_range = data["column_range"]

i = 5
ix = np.arange(ncol)

ylow, yhigh = 50, 50
ibeg, iend = 1500, 2000
ycen = np.polyval(orders[i], ix)
ycen_int = ycen.astype(int)
# yrange = extract.get_y_scale(ycen, [400, 600], [5, 5], nrow)

index = extract.make_index(ycen_int - ylow, ycen_int + yhigh, ibeg, iend)
swath_img = img[index]
swath_ycen = ycen[ibeg:iend] - ycen_int[ibeg:iend]
shear = 0
osample = 10

np.savetxt("image.txt", swath_img)
np.savetxt("ycen.txt", swath_ycen)

# return sp, sl, model, unc, mask
# data1 = cwrappers.slitfunc(swath_img, swath_ycen, osample=osample)
data2 = cwrappers.slitfunc_curved(swath_img, swath_ycen, shear, osample=osample)

data = data2
spec = data[0]
slitf = data[1]
model = data[2]
unc = data[3]
mask = data[4]

np.savetxt("spectrum.txt", spec)
np.savetxt("slitfunction.txt", slitf)
np.savetxt("model.txt", model)
np.savetxt("unc.txt", unc)
np.savetxt("mask.txt", mask)


# Input image
plt.subplot(121)
plt.imshow(swath_img, aspect="auto", origin="lower")
plt.xlabel("x")
plt.ylabel("y")

# Spectrum
plt.subplot(222)
x = np.indices(swath_img.shape)[1].ravel()

nsf = len(slitf)
nrow, ncol = swath_img.shape
sf = np.zeros(swath_img.shape)
for i in range(ncol):
    sf[:, i] = np.interp(
        np.linspace(0, nrow - 1, nrow),
        np.linspace(-1, nrow - 1 + 1, nsf) + (swath_ycen[i] - 0.5) * 1,
        slitf,
    )

    # plt.subplot(111)
    # plt.plot(swath_img[:, i] / np.mean(swath_img[:, i]), label="Image")
    # plt.plot(sf[:, i] / np.mean(sf[:, i]), label="Interpolated")
    # plt.plot(np.linspace(-1, nrow, nsf), slitf / np.mean(slitf), label="Slitfunction")
    # plt.legend()
    # plt.show()

# plt.subplot(122)
# plt.imshow(swath_img / sf, aspect="auto", origin="lower")
# plt.show()

y = swath_img / sf
y = y.ravel() * np.mean(spec) / np.mean(y)
plt.plot(x, y, ".", label="Observation")
plt.plot(spec, label="Extraction")
plt.title("Spectrum")
plt.xlabel("x [Pixel]")
plt.ylabel("Intensity [au]")
plt.legend()

# Slitfunction
plt.subplot(224)
x = np.indices(swath_img.shape)[0].ravel()
y = swath_img / spec[None, :]
y = y.ravel() * np.mean(slitf) / np.mean(y)

sf = np.interp(np.linspace(0, nrow - 1, nrow), np.linspace(-1, nrow, nsf), slitf)
plt.plot(x, y, ".", label="Observation")
plt.plot(sf, label="Extraction")
plt.title("Slitfunction")
plt.xlabel("y [Pixel]")
plt.ylabel("Intensity [au]")
plt.legend()

plt.show()

pass



import pytest
import numpy as np

from pyreduce import util
from pyreduce.extract import extract


def test_science(
    files,
    instrument,
    mode,
    mask,
    extension,
    bias,
    normflat,
    orders,
    settings,
    order_range,
):
    flat, blaze = normflat
    bias, _ = bias
    orders, column_range = orders
    settings = settings["science"]

    # Fix column ranges
    for i in range(blaze.shape[0]):
        column_range[i] = np.where(blaze[i] != 0)[0][[0, -1]]

    f = files["science"][0]

    im, head = util.load_fits(
        f, instrument, mode, extension, mask=mask, dtype=np.float32
    )
    # Correct for bias and flat field
    im -= bias
    im /= flat

    # Optimally extract science spectrum
    spec, sigma, _, _ = extract(
        im,
        orders,
        gain=head["e_gain"],
        readnoise=head["e_readn"],
        dark=head["e_drk"],
        column_range=column_range,
        order_range=order_range,
        extraction_type=settings["extraction_method"],
        extraction_width=settings["extraction_width"],
        lambda_sf=settings["smooth_slitfunction"],
        lambda_sp=settings["smooth_spectrum"],
        osample=settings["oversampling"],
        swath_width=settings["swath_width"],
        plot=False,
    )

    assert isinstance(spec, np.ma.masked_array)
    assert spec.ndim == 2
    assert spec.shape[0] == order_range[1] - order_range[0]
    assert spec.shape[1] == bias.shape[1]
    assert np.issubdtype(spec.dtype, np.floating)
    assert not np.any(np.isnan(spec))
    assert not np.all(np.all(spec.mask, axis=0))

    assert isinstance(sigma, np.ma.masked_array)
    assert sigma.ndim == 2
    assert sigma.shape[0] == order_range[1] - order_range[0]
    assert sigma.shape[1] == bias.shape[1]
    assert np.issubdtype(sigma.dtype, np.floating)
    assert not np.any(np.isnan(sigma))
    assert not np.all(np.all(sigma.mask, axis=0))



import pytest
import numpy as np
import os
import scipy.constants

from pyreduce import echelle

@pytest.fixture
def fname():
    name = "test_ech.ech"
    yield name

    if os.path.exists(name):
        os.remove(name)

@pytest.fixture
def clight():
    return scipy.constants.speed_of_light * 1e-3


def test_read_write_echelle(fname):
    data = np.linspace(1, 200, num=100, dtype=float)
    header = {"BLA": "Blub"}

    echelle.save(fname, header, test=data)

    ech = echelle.Echelle.read(fname)

    assert isinstance(ech, echelle.Echelle)
    assert "test" in ech
    assert np.allclose(ech["test"], data)

    assert ech.header["BLA"] == "Blub"

   

def test_expand_2d():
    # 2d make wave format
    wave = [0, 100, 2, 1, 0, 0, 0, 4, 2, 2, 1000, 100, 0, 100, 0, 0, 0, 0, 0]
    wave = np.array(wave)

    new = echelle.expand_polynomial(2, wave)

    assert isinstance(new, np.ndarray)
    assert new.ndim == 2
    assert new.shape[0] == 2
    assert new.shape[1] == 100

    cmp = np.linspace(1001, 1011, 100, endpoint=False)
    assert np.allclose(new[0], cmp)

    cmp = np.linspace(501, 506, 100, endpoint=False)
    assert np.allclose(new[1], cmp)


def test_expand_1d():
    wave = [[0.1, 1001], [0.05, 501]]
    wave = np.array(wave)

    new = echelle.expand_polynomial(100, wave)

    assert isinstance(new, np.ndarray)
    assert new.ndim == 2
    assert new.shape[0] == 2
    assert new.shape[1] == 100

    cmp = np.linspace(1001, 1011, 100, endpoint=False)
    assert np.allclose(new[0], cmp)

    cmp = np.linspace(501, 506, 100, endpoint=False)
    assert np.allclose(new[1], cmp)

def test_wavelength_correction(fname, clight):
    bc = 100
    rv = 200
    header = {"barycorr": bc, "radvel": rv}
    wave = np.linspace(5000, 6000, 100)
    wave = np.atleast_2d(wave)

    echelle.save(fname, header, wave=wave)

    # Read raw frame
    ech = echelle.Echelle.read(fname, raw=True)
    assert np.allclose(ech["wave"], wave)
    assert ech.header["barycorr"] == bc
    assert ech.header["radvel"] == rv

    # Apply only barycentric correction
    ech = echelle.Echelle.read(fname, barycentric_correction=True, radial_velociy_correction=False)
    assert np.allclose(ech["wave"], wave * (1 + bc / clight))
    assert ech.header["barycorr"] == 0
    assert ech.header["radvel"] == rv


    # Apply only radial velocity correction
    ech = echelle.Echelle.read(fname, barycentric_correction=False, radial_velociy_correction=True)
    assert np.allclose(ech["wave"], wave * (1 - rv / clight))
    assert ech.header["barycorr"] == bc
    assert ech.header["radvel"] == 0

    # Apply both corrections
    ech = echelle.Echelle.read(fname, barycentric_correction=True, radial_velociy_correction=True)
    assert np.allclose(ech["wave"], wave * (1 + (bc - rv) / clight))
    assert ech.header["barycorr"] == 0
    assert ech.header["radvel"] == 0

def test_column_range_mask(fname):
    spec = np.linspace(1, 1000, 100)
    spec = np.stack([spec, spec])

    columns = np.array([[10, 99],[0, 50]])

    echelle.save(fname, {}, spec=spec, columns=columns)

    ech = echelle.read(fname)

    assert "mask" in ech
    assert isinstance(ech["spec"], np.ma.masked_array)

    for iord in range(2):
        assert all(ech["mask"][iord, :columns[iord, 0]])
        assert all(ech["mask"][iord, columns[iord, 1]:])
        assert not any(ech["mask"][iord, columns[iord, 0]:columns[iord, 0]])

def test_continuum_normalization(fname):
    spec = np.full((2, 100), 10, dtype=float)
    sig = np.full((2, 100), 1, dtype=float)
    cont = np.full((2, 100), 2, dtype=float)

    echelle.save(fname, {}, spec=spec, sig=sig, cont=cont)
    ech = echelle.read(fname)

    assert np.allclose(ech["spec"], spec/cont)
    assert np.allclose(ech["sig"], sig/cont)
    assert np.allclose(ech["cont"], cont)




import pytest

from pyreduce import reduce


def test_configuration():
    config = reduce.load_config(None, "UVES", 0)
    assert isinstance(config, dict)

    config = reduce.load_config("settings_UVES.json", "UVES", 0)
    assert isinstance(config, dict)

    config = reduce.load_config({"UVES": "settings_UVES.json"}, "UVES", 0)
    assert isinstance(config, dict)

    config = reduce.load_config(["settings_UVES.json"], "UVES", 0)
    assert isinstance(config, dict)

    with pytest.raises(KeyError):
        config = reduce.load_config({"UVES": "settings_UVES.json"}, "HARPS", 0)

    with pytest.raises(IndexError):
        config = reduce.load_config(["settings_UVES.json"], "UVES", 1)


def test_main():
    reduce.main(steps=())



import pytest

from os.path import dirname, join
import astropy.io.fits as fits
import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import gaussian

# from skimage import transform as tf

# from pyreduce.clib.build_extract import build

# build()

from pyreduce import extract
from pyreduce import util
from pyreduce.cwrappers import slitfunc


# @pytest.fixture
# def testdata1():
#     folder = dirname(__file__)
#     fname = "test.dat"
#     return join(folder, fname)


# @pytest.fixture
# def testdata2():
#     folder = dirname(__file__)
#     fname = "test2.dat"
#     return join(folder, fname)


# @pytest.fixture
# def testdata3():
#     folder = dirname(__file__)
#     fname = "test3.dat"
#     return join(folder, fname)


# @pytest.fixture
# def testdata2_after():
#     folder = dirname(__file__)
#     fname = "test2_after.dat"
#     return join(folder, fname)


@pytest.fixture
def width():
    return 100


@pytest.fixture
def height():
    return 50


@pytest.fixture(params=["linear", "sine"])
def spec(request, width):
    name = request.param

    if name == "linear":
        return 5 + np.linspace(0, 5, width)
    if name == "sine":
        return 5 + np.sin(np.linspace(0, 20 * np.pi, width))


@pytest.fixture(params=["gaussian"])
def slitf(request, height, oversample):
    name = request.param

    if name == "gaussian":
        y = gaussian(height * oversample, height / 8 * oversample)
        return y / np.sum(y)
    if name == "rectangular":
        x = np.arange(height * oversample)
        y = np.where(
            (x > height * oversample / 4) & (x < height * oversample * 3 / 4), 1, 0
        )
        return y / np.sum(y)


@pytest.fixture(params=[1, 2, 10], ids=["os=1", "os=2", "os=10"])
def oversample(request):
    return request.param


@pytest.fixture(params=[0], ids=["noise=0"])
def noise(request):
    return request.param


@pytest.fixture(params=["straight"], ids=["ycen=straight"])
def ycen(request, width, height):
    name = request.param
    if name == "straight":
        return np.full(width, (height - 1) / 2)

    if name == "linear":
        delta = 2
        return np.linspace(height / 2 - delta / 2, height / 2 + delta / 2, num=width)


@pytest.fixture
def orders(width, ycen):
    fit = np.polyfit(np.arange(width), ycen, deg=5)
    return np.atleast_2d(fit)
    # return np.array([[(height - 1) / 2]])


@pytest.fixture
def sample_data(width, height, spec, slitf, oversample, noise, ycen, tilt=0):
    img = spec[None, :] * slitf[:, None]
    img += noise * np.random.randn(*img.shape)

    # TODO more sophisticated sample data creation

    # afine_tf = tf.AffineTransform(shear=-shear)
    # img = tf.warp(img, inverse_map=afine_tf)

    # # apply order curvature
    # big_height = (int(np.ceil(np.max(ycen))) + height) * oversample
    # big_height += oversample - (big_height % oversample)
    # big_img = np.zeros((big_height, width))
    # index = util.make_index(
    #     (oversample * ycen).astype(int) - height // 2 * oversample,
    #     (ycen * oversample).astype(int) + height // 2 * oversample - 1,
    #     0,
    #     width,
    # )
    # big_img[index] = img

    # # downsample oversampling
    # img = big_img[::oversample]
    # for i in range(1, oversample):
    #     img += big_img[i::oversample]
    # img /= oversample

    return img, spec, slitf


def test_extend_orders():
    # Test normal case
    orders = np.array([[0.1, 5], [0.1, 7]])
    extended = extract.extend_orders(orders, 10)

    assert np.array_equal(orders, extended[1:-1])
    assert np.array_equal(extended[0], [0.1, 3])
    assert np.array_equal(extended[-1], [0.1, 9])

    # Test just one order
    orders = np.array([0.1, 5], ndmin=2)
    extended = extract.extend_orders(orders, 10)

    assert np.array_equal(orders, extended[1:-1])
    assert np.array_equal(extended[0], [0, 0])
    assert np.array_equal(extended[-1], [0, 10])


def test_fix_column_range():
    # Some orders will be shortened
    img = np.zeros((50, 1000))
    orders = np.array([[0.2, 3], [0.2, 5], [0.2, 7], [0.2, 9]])
    ew = np.array([[10, 10], [10, 10], [10, 10], [10, 10]])
    cr = np.array([[0, 1000], [0, 1000], [0, 1000], [0, 1000]])

    fixed = extract.fix_column_range(img, orders, ew, cr)

    assert np.array_equal(fixed[1], [25, 175])
    assert np.array_equal(fixed[2], [15, 165])
    assert np.array_equal(fixed[0], fixed[1])
    assert np.array_equal(fixed[-1], fixed[-1])

    # Nothing should change here
    orders = np.array([[20], [20], [20]])
    ew = np.array([[10, 10], [10, 10], [10, 10]])
    cr = np.array([[0, 1000], [0, 1000], [0, 1000]])

    fixed = extract.fix_column_range(img, orders, ew, np.copy(cr))
    assert np.array_equal(fixed, cr)


def test_arc_extraction(sample_data, orders, width, noise, oversample):
    img, spec, slitf = sample_data

    # orders = np.array([orders[0], orders[0], orders[0]])
    extraction_width = np.array([[10, 10]])
    column_range = np.array([[0, width]])
    # column_range = extract.fix_column_range(img, orders, extraction_width, column_range)

    spec_out, unc_out = extract.arc_extraction(
        img, orders, extraction_width, column_range
    )

    assert isinstance(spec_out, np.ndarray)
    assert spec_out.ndim == 2
    assert spec_out.shape[0] == 1
    assert spec_out.shape[1] == width

    assert isinstance(unc_out, np.ndarray)
    assert unc_out.ndim == 2
    assert unc_out.shape[0] == 1
    assert unc_out.shape[1] == width

    assert np.abs(np.diff(spec_out / spec)).max() < noise * 10 + 1e-8
    assert np.abs(np.diff(unc_out / spec_out)).max() < oversample / 5 + 1e-1


def test_vertical_extraction(sample_data, orders, width, height, noise, oversample):
    img, spec, slitf = sample_data

    spec_vert, sunc_vert, slitf_vert, _ = extract.extract(img, orders)

    assert isinstance(spec_vert, np.ma.masked_array)
    assert spec_vert.ndim == 2
    assert spec_vert.shape[0] == orders.shape[0]
    assert spec_vert.shape[1] == width

    assert isinstance(sunc_vert, np.ma.masked_array)
    assert sunc_vert.ndim == 2
    assert sunc_vert.shape[0] == orders.shape[0]
    assert sunc_vert.shape[1] == width

    assert isinstance(slitf_vert, np.ndarray)
    assert slitf_vert.ndim == 2
    assert slitf_vert.shape[0] == orders.shape[0]
    assert slitf_vert.shape[1] <= height * oversample

    assert not np.any(spec_vert == 0)
    assert np.abs(np.diff(spec / spec_vert[0])).max() <= noise + 1e-1

    assert not np.any(sunc_vert == 0)
    assert np.abs(sunc_vert / spec_vert).max() <= noise * 1.1 * oversample + 1e-2

    # cut = (height - slitf_vert.shape[1]) / 2
    # cut = (int(np.floor(cut)), int(np.ceil(cut)))
    # xnew = np.linspace(cut[0] + 0.5, cut[1] + 0.5, slitf_vert.shape[1])
    # cutout = np.interp(xnew, np.arange(0, height), slitf)
    # assert cutout.shape[0] == slitf_vert.shape[1]
    # assert cutout / slitf == 1


def test_curved_equal_vertical_extraction(sample_data, orders, noise):
    # Currently extract always uses the vertical extraction, making this kind of useless
    img, spec, slitf = sample_data
    tilt = 0
    shear = 0

    spec_curved, sunc_curved, slitf_curved, _ = extract.extract(
        img, orders, tilt=tilt, shear=shear
    )
    spec_vert, sunc_vert, slitf_vert, _ = extract.extract(img, orders)

    assert np.allclose(spec_curved, spec_vert, rtol=1e-2)
    # assert np.allclose(sunc_curved, sunc_vert, rtol=0.1)
    assert np.allclose(slitf_curved, slitf_vert, rtol=1e-1)



"""
Abstract parent module for all other instruments
Contains some general functionality, which may be overridden by the children of course
"""
import os.path
import datetime
import glob
import logging
import json

import numpy as np
from astropy.io import fits
from astropy.time import Time
from dateutil import parser


def find_first_index(arr, value):
    """ find the first element equal to value in the array arr """
    try:
        return next(i for i, v in enumerate(arr) if v == value)
    except StopIteration:
        raise Exception("Value %s not found" % value)


def observation_date_to_night(observation_date):
    """Convert an observation timestamp into the date of the observation night
    Nights start at 12am and end at 12 am the next day

    Parameters
    ----------
    observation_date : datetime
        timestamp of the observation

    Returns
    -------
    night : datetime.date
        night of the observation
    """
    if observation_date == "":
        return None

    observation_date = parser.parse(observation_date)
    oneday = datetime.timedelta(days=1)

    if observation_date.hour < 12:
        observation_date -= oneday
    return observation_date.date()


class getter:
    """Get data from a header/dict, based on the given mode, and applies replacements"""

    def __init__(self, header, info, mode):
        self.header = header
        self.info = info
        try:
            self.index = find_first_index(info["modes"], mode)
        except KeyError:
            logging.warning("No instrument modes found in instrument info")
            self.index = 0

        # Pick values for the given mode
        for k, v in self.info.items():
            if isinstance(v, list):
                self.info[k] = v[self.index]

    def __call__(self, key, alt=None):
        return self.get(key, alt)

    def get(self, key, alt=None):
        """Get data

        Parameters
        ----------
        key : str
            key of the data in the header
        alt : obj, optional
            alternative value, if key does not exist (default: None)

        Returns
        -------
        value : obj
            value found in header (or alternatively alt)
        """

        value = self.info.get(key, key)
        if isinstance(value, list):
            value = value[self.index]
        if isinstance(value, str):
            value = value.format(**self.info)
            value = self.header.get(value, alt)
        return value


class instrument:
    """
    Abstract parent class for all instruments
    Handles the handling of instrument specific information
    """

    def __init__(self):
        self.instrument = self.__class__.__name__.lower()
        # raise NotImplementedError("This is an abstract class")

    def load_info(self):
        """
        Load static instrument information
        Either as fits header keywords or static values

        Returns
        ------
        info : dict(str:object)
            dictionary of REDUCE names for properties to Header keywords/static values
        """
        # Tips & Tricks:
        # if several modes are supported, use a list for modes
        # if a value changes depending on the mode, use a list with the same order as "modes"
        # you can also use values from this dictionary as placeholders using {name}, just like str.format

        this = os.path.dirname(__file__)
        fname = f"{self.instrument.lower()}.json"
        fname = os.path.join(this, fname)
        with open(fname) as f:
            info = json.load(f)
        return info

    def add_header_info(self, header, mode, **kwargs):
        """read data from header and add it as REDUCE keyword back to the header

        Parameters
        ----------
        header : fits.header, dict
            header to read/write info from/to
        mode : str
            instrument mode

        Returns
        -------
        header : fits.header, dict
            header with added information
        """

        info = self.load_info()
        get = getter(header, info, mode)

        header["INSTRUME"] = get("instrument", self.__class__.__name__)
        header["TELESCOP"] = get("telescope", "")
        header["EXPTIME"] = get("exposure_time", 0)
        header["MJD-OBS"] = get("date", 0)

        if isinstance(header["MJD-OBS"], str):
            try:
                header["MJD-OBS"] = Time(header["MJD-OBS"]).mjd
            except:
                logging.warning("Unable to determine the MJD date of the observation")

        header["e_orient"] = get("orientation", 0)

        naxis_x = get("naxis_x", 0)
        naxis_y = get("naxis_y", 0)

        prescan_x = get("prescan_x", 0)
        overscan_x = get("overscan_x", 0)
        prescan_y = get("prescan_y", 0)
        overscan_y = get("overscan_y", 0)

        header["e_xlo"] = prescan_x
        header["e_xhi"] = naxis_x - overscan_x

        header["e_ylo"] = prescan_y
        header["e_yhi"] = naxis_y - overscan_y

        header["e_gain"] = get("gain", 1)
        header["e_readn"] = get("readnoise", 0)

        header["e_sky"] = get("sky", 0)
        header["e_drk"] = get("dark", 0)
        header["e_backg"] = header["e_gain"] * (header["e_drk"] + header["e_sky"])

        header["e_imtype"] = get("image_type")
        header["e_ctg"] = get("category")

        header["e_ra"] = get("ra", 0)
        header["e_dec"] = get("dec", 0)
        header["e_jd"] = get("jd", 0)

        header["e_obslon"] = get("longitude")
        header["e_obslat"] = get("latitude")
        header["e_obsalt"] = get("altitude")

        header["e_wavecal_element"] = get("wavecal_element", info.get("wavecal_element", None))
        return header

    def sort_files(self, input_dir, target, night, mode, **kwargs):
        """
        Sort a set of fits files into different categories
        types are: bias, flat, wavecal, orderdef, spec

        Parameters
        ----------
        input_dir : str
            input directory containing the files to sort
        target : str
            name of the target as in the fits headers
        night : str
            observation night, possibly with wildcards
        mode : str
            instrument mode
        Returns
        -------
        files_per_night : list[dict{str:dict{str:list[str]}}]
            a list of file sets, one entry per night, where each night consists of a dictionary with one entry per setting,
            each fileset has five lists of filenames: "bias", "flat", "order", "wave", "spec", organised in another dict
        nights_out : list[datetime]
            a list of observation times, same order as files_per_night
        """

        # TODO allow several names for the target?

        info = self.load_info()
        target = target.upper()
        instrument = info.get("__instrument__", "").upper()

        # Try matching with nights
        try:
            night = parser.parse(night).date()
            individual_nights = [night]
        except ValueError:
            # if the input night can't be parsed, use all nights
            # Usually the case if wildcards are involved
            individual_nights = "all"

        # find all fits files in the input dir(s)
        input_dir = input_dir.format(
            instrument=instrument.upper(), target=target, mode=mode, night=night
        )
        files = glob.glob(input_dir + "/*.fits")
        files += glob.glob(input_dir + "/*.fits.gz")
        files = np.array(files)

        # Load the mode identifier for the current mode from the header
        # This could be anything really, e.g. the size of the data axis
        i = [i for i, m in enumerate(info["modes"]) if m == mode][0]
        mode_id = info["modes_id"][i].upper()

        # Initialize arrays
        # observed object
        ob = np.zeros(len(files), dtype="U20")
        # observation type, bias, flat, spec, etc.
        ty = np.zeros(len(files), dtype="U20")
        # instrument mode, e.g. red, blue
        mo = np.zeros(len(files), dtype="U20")
        # special setting identifier, e.g. wavelength setting
        se = np.zeros(len(files), dtype="U20")
        # observed night, parsed into a datetime object
        ni = np.zeros(len(files), dtype=datetime.datetime)
        # instrument, used for observation
        it = np.zeros(len(files), dtype="U20")

        for i, f in enumerate(files):
            h = fits.open(f)[0].header
            ob[i] = h.get(info["target"], "")
            ty[i] = h.get(info["observation_type"], "")
            # The mode descriptor has different names in different files, so try different ids
            mo[i] = h.get(info["instrument_mode"])
            ni_tmp = h.get(info["date"], "")
            it[i] = h.get(info["instrument"], "")
            se[i] = h.get(info["setting_specifier"], "")

            # Sanitize input
            ni[i] = observation_date_to_night(ni_tmp)
            ob[i] = ob[i].replace("-", "")

        if isinstance(individual_nights, str) and individual_nights == "all":
            individual_nights = np.unique(ni)
            logging.info(
                "Can't parse night %s, use all %i individual nights instead",
                night,
                len(individual_nights),
            )

        files_per_night = []
        nights_out = []
        for ind_night in individual_nights:
            # Select files for this night, this instrument, this instrument mode
            selection = (ni == ind_night) & (it == instrument) & (mo == mode_id)

            # Find all unique setting keys for this night and target
            # Only look at the settings of observation files
            keys = se[(ty == info["id_spec"]) & (ob == target) & selection]
            keys = np.unique(keys)

            files_this_night = {}
            for key in keys:
                select = selection & (se == key)

                # find all relevant files for this setting
                # bias ignores the setting
                files_this_night[key] = {
                    "bias": files[(ty == info["id_bias"]) & selection],
                    "flat": files[(ty == info["id_flat"]) & select],
                    "orders": files[(ty == info["id_orders"]) & select],
                    "wavecal": files[(ob == info["id_wave"]) & select],
                    "science": files[(ty == info["id_spec"]) & (ob == target) & select],
                }

            if len(keys) != 0:
                nights_out.append(ind_night)
                files_per_night.append(files_this_night)

        return files_per_night, nights_out

    def get_wavecal_filename(self, header, mode, **kwargs):
        """Get the filename of the pre-existing wavelength solution for the current setting

        Parameters
        ----------
        header : fits.header, dict
            header of the wavelength calibration file
        mode : str
            instrument mode

        Returns
        -------
        filename : str
            name of the wavelength solution file
        """

        info = self.load_info()
        specifier = header.get(info.get("wavecal_specifier", ""), "")
        instrument = "wavecal"

        cwd = os.path.dirname(__file__)
        fname = "{instrument}_{mode}_{specifier}.npz".format(
            instrument=instrument.lower(), mode=mode, specifier=specifier
        )
        fname = os.path.join(cwd, "..", "wavecal", fname)
        return fname

    def get_supported_modes(self):
        info = self.load_info()
        return info["modes"]


class COMMON(instrument):
    def load_info(self):
        return {
            "naxis_x": "NAXIS1",
            "naxis_y": "NAXIS2",
            "modes": [""],
            "modes_id": [""],
        }




"""
Handles instrument specific info for the HARPS spectrograph

Mostly reading data from the header
"""
import os.path
import glob
import logging
from datetime import datetime
import fnmatch
import json

import numpy as np
from astropy.io import fits
from dateutil import parser

from .common import getter, instrument, observation_date_to_night


class JWST_MIRI(instrument):

    def add_header_info(self, header, mode, **kwargs):
        """ read data from header and add it as REDUCE keyword back to the header """
        # "Normal" stuff is handled by the general version, specific changes to values happen here
        # alternatively you can implement all of it here, whatever works
        header = super().add_header_info(header, mode)
        info = self.load_info()
        return header

    def sort_files(self, input_dir, target, night, mode):
        """
        Sort a set of fits files into different categories
        types are: bias, flat, wavecal, orderdef, spec

        Parameters
        ----------
        input_dir : str
            input directory containing the files to sort
        target : str
            name of the target as in the fits headers
        night : str
            observation night, possibly with wildcards
        mode : str
            instrument mode
        Returns
        -------
        files_per_night : list[dict{str:dict{str:list[str]}}]
            a list of file sets, one entry per night, where each night consists of a dictionary with one entry per setting,
            each fileset has five lists of filenames: "bias", "flat", "order", "wave", "spec", organised in another dict
        nights_out : list[datetime]
            a list of observation times, same order as files_per_night
        """

        info = self.load_info()
        target = target.upper()
        instrument = "MIRI"

        # Try matching with nights
        try:
            night = parser.parse(night).date()
            individual_nights = [night]
        except ValueError:
            # if the input night can't be parsed, use all nights
            # Usually the case if wildcards are involved
            individual_nights = "all"

        # find all fits files in the input dir(s)
        input_dir = input_dir.format(
            instrument=self.instrument.upper(), target=target, mode=mode, night=night
        )
        files = glob.glob(input_dir + "/*.fits")
        files += glob.glob(input_dir + "/*.fits.gz")
        files = np.array(files)

        # Load the mode identifier for the current mode from the header
        # This could be anything really, e.g. the size of the data axis
        i = [i for i, m in enumerate(info["modes"]) if m == mode][0]
        mode_id = info["modes_id"][i].upper()

        # Initialize arrays
        # observed object
        ob = np.zeros(len(files), dtype="U20")
        # observation type, bias, flat, spec, etc.
        ty = np.zeros(len(files), dtype="U20")
        # instrument mode, e.g. red, blue
        mo = np.zeros(len(files), dtype="U20")
        # special setting identifier, e.g. wavelength setting
        se = np.zeros(len(files), dtype="U20")
        # observed night, parsed into a datetime object
        ni = np.zeros(len(files), dtype=datetime)
        # instrument, used for observation
        it = np.zeros(len(files), dtype="U20")

        for i, f in enumerate(files):
            h = fits.open(f)[0].header
            it[i] = h.get(info["instrument"], "")
            ty[i] = h.get(info["observation_type"], "")
            # The mode descriptor has different names in different files, so try different ids
            mo[i] = h.get(info["instrument_mode"])
            ni_tmp = h.get(info["date"], "") + "T" + h.get(info["time"], "")
            se[i] = instrument

            # Sanitize input
            ni[i] = observation_date_to_night(ni_tmp)
            ob[i] = ob[i].replace("-", "")

        if isinstance(individual_nights, str) and individual_nights == "all":
            individual_nights = np.unique(ni)
            logging.info(
                "Can't parse night %s, use all %i individual nights instead",
                night,
                len(individual_nights),
            )

        files_per_night = []
        nights_out = []
        for ind_night in individual_nights:
            # Select files for this night, this instrument, this instrument mode
            selection = (ni == ind_night) & (it == instrument) & (mo == mode_id)

            # Find all unique setting keys for this night and target
            # Only look at the settings of observation files
            match_ty = np.array([fnmatch.fnmatch(t, info["id_spec"]) for t in ty])

            keys = se[match_ty & selection]
            keys = np.unique(keys)

            files_this_night = {}
            for key in keys:
                select = selection & (se == key)

                # find all relevant files for this setting
                # bias ignores the setting
                files_this_night[key] = {
                    "bias": files[(ty == info["id_bias"]) & selection],
                    "flat": files[(ty == info["id_flat"]) & select],
                    "orders": files[(ty == info["id_flat"]) & select],
                    "wavecal": files[(ty == info["id_wave"]) & select],
                    "science": files[match_ty & select],
                }
                # TODO find actual flat files
                files_this_night[key]["flat"] = [files_this_night[key]["science"][0]]
                files_this_night[key]["orders"] = [files_this_night[key]["science"][0]]
                files_this_night[key]["curvature"] = files_this_night[key]["wavecal"]
                

            if len(keys) != 0:
                nights_out.append(ind_night)
                files_per_night.append(files_this_night)

        return files_per_night, nights_out

    def get_wavecal_filename(self, header, mode, **kwargs):
        """ Get the filename of the wavelength calibration config file """
        cwd = os.path.dirname(__file__)
        fname = "{instrument}_{mode}_2D.npz".format(instrument="harps", mode=mode)
        fname = os.path.join(cwd, "..", "wavecal", fname)
        return fname



"""
Simple usage example for PyReduce
Loads a sample UVES dataset, and runs the full extraction
"""

import os.path
import pyreduce
from pyreduce import datasets


# define parameters
instrument = "NIRSPEC"
target = "GJ1214"
night = "all"
mode = "NIRSPEC"
steps = (
    "bias",
    "flat",
    "orders",
    "norm_flat",
    "wavecal",
    "freq_comb",
    # "curvature",
    "science",
    # "continuum",
    "finalize",
)

# some basic settings
# Expected Folder Structure: base_dir/datasets/HD132205/*.fits.gz
# Feel free to change this to your own preference, values in curly brackets will be replaced with the actual values {}

# load dataset (and save the location)
base_dir = "/DATA/Keck/{instrument}/GJ1214_b/"
input_dir = "raw/sci/"
output_dir = "reduced"

# Path to the configuration parameters, that are to be used for this reduction
config = os.path.join(os.path.dirname(__file__), "settings_NIRSPEC.json")

pyreduce.reduce.main(
    instrument,
    target,
    night,
    mode,
    steps,
    base_dir=base_dir,
    input_dir=input_dir,
    output_dir=output_dir,
    configuration=config,
    # order_range=(0, 25),
)



"""
Module that
a) Clips: remove pre- and overscan regions
and b) Flips: orients the image so that orders are roughly horizontal
"""

import numpy as np


def clipnflip(image, header, xrange=None, yrange=None, orientation=None):
    """
    Process an image and associated FITS header already in memory as follows:
    1. Trim image to desired subregion: newimage = image(xlo:xhi,ylo:yhi)
    2. Transform to standard orientation (red at top, orders run left to right)

    Parameters
    -----------
    image : array[nrow, ncol]
        raw image to be processed.
    header : fits.header, dict
        FITS header of the image
    xrange : tuple(int, int), optional
        column - range to keep in the image (default: data from header/instrument)
    yrange : tuple(int, int), optional
        row - range to keep in the image (default: data from header/instrument)
    orientation : int, optional
        number of counterclockwise 90 degrees rotation to apply to the image (default: data from header/instrument)

    Returns
    -------
    image : array[yrange, xrange]
        clipped and flipped image

    Raises
    -------
    NotImplementedError
        nonlinear images are not supported yet
    """

    # This part depends on how many amplifiers were used for the readout
    n_amp = header.get("e_ampl", 1)

    if n_amp > 1:  # more than one amplifier
        xlo = np.array(header["e_xlo*"].values())
        xhi = np.array(header["e_xhi*"].values())
        ylo = np.array(header["e_ylo*"].values())
        yhi = np.array(header["e_yhi*"].values())

        # Make sure trim region is a subset of actual image.
        sz = image.shape
        if (
            np.any(xlo < 0)
            | np.any(xlo >= sz[1])
            | np.any(ylo < 0)
            | np.any(ylo >= sz[0])
            | np.any(xhi < 0)
            | np.any(xhi > sz[1])
            | np.any(yhi < 0)
            | np.any(yhi > sz[0])
        ):
            raise ValueError("Error specifying trim region")

        linear = header.get("e_linear", False)
        if not linear:
            # TODO
            raise NotImplementedError("only linear for now")
            # pref = header["e_prefmo"]
            # image = call_function("nonlinear_" + pref, image, header)

            # i = np.where(header["e_linear"] >= 0)
            # if len(i) > 0:
            #     header = header[0 : i - 1 + 1] + header[i + 1 :]
            # header["e_linear"] = ("t", "image corrected of non-linearity")
            # ii = np.where(header["e_gain*"] >= 0)
            # if len(ii) > 0:
            #     for i in np.arange(0, len(ii) - 1 + 1, 1):
            #         k = ii[i]
            #         header = [header[0 : k - 1 + 1], header[k + 1 :]]
            # header["e_gain"] = (1, "image was converted to e-")

        # Trim image to leave only the subimage containing valid image data.
        # For two amplifiers we assume a single vertical or horizontal gap.
        # With four amplifiers we can have a cross.

        if n_amp == 2:
            # TODO this needs testing
            if xlo[0] == xlo[1]:
                xsize = xhi[0] - xlo[0]
                ysize = yhi[0] - ylo[0] + yhi[1] - ylo[1]
                timage = np.empty((xsize, ysize), dtype=image.dtype)
                ysize = yhi[0] - ylo[0]
                timage[:ysize, :xsize] = image[xlo[0]]
                timage[ysize:, :xsize] = image[xlo[1]]
            elif ylo[0] == ylo[1]:
                xsize = xhi[0] - xlo[0] + xhi[1] - xlo[1]
                ysize = yhi[0] - ylo[0]
                timage = np.empty((xsize, ysize), dtype=image.dtype)
                xsize = xhi[0] - xlo[0]
                timage[:ysize, :xsize] = image[xlo[0]]
                timage[:ysize, xsize:] = image[xlo[1]]
            else:
                raise Exception(
                    "The two ccd sections are aligned neither in x nor in y"
                )
        elif n_amp == 4:
            raise NotImplementedError("4-amplifier section is not implemented yet")
    else:
        xlo, xhi = xrange if xrange is not None else (header["e_xlo"], header["e_xhi"])
        ylo, yhi = yrange if yrange is not None else (header["e_ylo"], header["e_yhi"])

        # Make sure trim region is a subset of actual image.
        sz = image.shape[-2:]
        if not (0 <= xlo < xhi <= sz[1] and 0 <= ylo < yhi <= sz[0]):
            raise IndexError(
                "Image Clipping Indices are not within the image (or in inverse order)"
            )

        # Trim image to leave only the subimage containing valid image data.
        timage = image[..., ylo:yhi, xlo:xhi]  # trimmed image

    # Flip image (if necessary) to achieve standard image orientation.
    orientation = orientation if orientation is not None else header.get("e_orient")
    if orientation is not None:
        timage = np.rot90(timage, -1 * orientation, axes=(-2, -1))

    # TODO just sum up groups and stuff?
    while timage.ndim > 2:
        timage = np.sum(timage, axis=0)

    return timage



from os.path import dirname, join, basename, exists
from glob import glob

import pytest

from pyreduce.configuration import get_configuration_for_instrument
from pyreduce.instruments import instrument_info
from pyreduce.instruments import common


supported_instruments = glob(join(dirname(__file__), "../pyreduce/instruments/*.json"))
supported_instruments = [basename(f)[:-5] for f in supported_instruments]
supported_instruments = [f for f in supported_instruments if f not in ["common"]]


@pytest.fixture(params=supported_instruments)
def supported_instrument(request):
    return request.param


@pytest.fixture
def supported_modes(supported_instrument):
    return instrument_info.get_supported_modes(supported_instrument)

@pytest.fixture
def config(supported_instrument):
    return get_configuration_for_instrument(supported_instrument)


def test_load_common():
    instr = instrument_info.load_instrument(None)
    assert isinstance(instr, common.instrument)
    assert isinstance(instr, common.COMMON)


def test_load_instrument(supported_instrument):
    instr = instrument_info.load_instrument(supported_instrument)
    assert isinstance(instr, common.instrument)


def test_get_instrument_info(supported_instrument):
    info = instrument_info.get_instrument_info(supported_instrument)
    assert isinstance(info, dict)


def test_modeinfo(supported_instrument, supported_modes):
    # Standard FITS header keywords
    required_keywords = ["INSTRUME", "TELESCOP", "EXPTIME", "MJD-OBS"]
    # PyReduce keywords
    required_keywords += [
        "e_xlo",
        "e_xhi",
        "e_ylo",
        "e_yhi",
        "e_gain",
        "e_readn",
        "e_sky",
        "e_drk",
        "e_backg",
        "e_imtype",
        "e_ctg",
        "e_ra",
        "e_dec",
        "e_jd",
        "e_obslon",
        "e_obslat",
        "e_obsalt",
    ]
    for mode in supported_modes:
        header = instrument_info.modeinfo({}, supported_instrument, mode)
        assert isinstance(header, dict)
        for key in required_keywords:
            assert key in header.keys()


def test_sort_files(supported_instrument, supported_modes, config):
    for mode in supported_modes:
        files, nights = instrument_info.sort_files(
            ".", "", "", supported_instrument, mode, **config["instrument"]
        )
        assert isinstance(files, list)
        assert isinstance(nights, list)
        assert len(files) == len(nights)

        # TODO Test contents of the lists
        # Maybe create debug files for each instrument and then make sure only the correct ones are in each list?
        # That would require a function for each instrument that creates the appropiate files
        # if len(files) != 0:
        # f = files[0][list(files[0].keys())[0]]
        # assert "bias" in f.keys()
        # assert "flat" in f.keys()
        # assert "wavecal" in f.keys()
        # assert "curvature" in f.keys()
        # assert "orders" in f.keys()
        # assert "science" in f.keys()

@pytest.mark.skip(reason="No wavelength calibration files for most instruments present at the moment")
def test_get_wavecal_name(supported_instrument, supported_modes):
    for mode in supported_modes:
        wname = instrument_info.get_wavecal_filename({}, supported_instrument, mode)
        assert isinstance(wname, str)
        assert exists(wname)




import pytest

import astropy.io.fits as fits
import numpy as np
import tempfile
import os

from pyreduce import combine_frames


@pytest.fixture
def instrument():
    return "UVES"


@pytest.fixture
def mode():
    return "middle"


@pytest.fixture
def files():
    n = 3
    files = [
        tempfile.NamedTemporaryFile(suffix=".fits", delete=False) for _ in range(n)
    ]
    files = [f.name for f in files]
    yield files
    # Tear down code
    for f in files:
        os.remove(f)


def test_running_median():
    arr = np.array([np.arange(10), np.arange(1, 11), np.arange(2, 12)])
    size = 3

    result = combine_frames.running_median(arr, size)
    compare = np.array([np.arange(1, 9), np.arange(2, 10), np.arange(3, 11)])

    assert np.array_equal(result, compare)


def test_running_mean():
    arr = np.array([np.arange(10), np.arange(1, 11), np.arange(2, 12)])
    size = 3

    result = combine_frames.running_median(arr, size)
    compare = np.array([np.arange(1, 9), np.arange(2, 10), np.arange(3, 11)])

    assert np.array_equal(result, compare)


def test_combine_frames(files):
    img = np.full((100, 100), 10)
    ovscx = 5
    for f in files:
        data = img + np.random.randint(0, 20, size=img.shape)
        head = fits.Header(
            cards={
                "ESO DET OUT1 PRSCX": 0,
                "ESO DET OUT1 OVSCX": ovscx,
                "ESO DET OUT1 CONAD": 1,
                "ESO DET OUT1 RON": 0,
                "EXPTIME": 10,
                "RA": 100,
                "DEC": 51,
                "MJD-OBS": 12030,
            }
        )
        fits.writeto(f, data=data, header=head)

    combine, chead = combine_frames.combine_frames(files, "UVES", "middle", 0, window=5)

    assert combine.shape[0] == img.shape[0] - ovscx
    assert combine.shape[1] == img.shape[1]

    assert chead["exptime"] == 10 * len(files)

